# sdi.py Sputter Deposition Investigator
# module to monitor and control sputter deposition
# derived from measure/spectroscopy.py by AS@KIT and others
# YS,JNV,HR@KIT 05/2018 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from __future__ import print_function

import qkit

from qkit.measure.measurement_class import Measurement
import qkit.measure.write_additional_files as waf
from qkit.storage import store as hdf
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.gui.notebook.Stopwatch import Stopwatch
from qkit.gui.plot import plot as qviewkit

import sys
import time
import threading
from threading import Thread

try:  # In Python 2 the following command works only with python-future installed
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

import numpy as np
from scipy.optimize import curve_fit


class SputterMonitor(object):
    """
    Sputter monitoring class, derived from spectroscopy.py
    monitors and stores sputter parameters live during deposition
    facilitates dynamic flow adjustment

    Example:
        m = SPUTTER_Monitor(quartz, ohmmeter, film_name, sample)
        m.set_filmparameters() ... see in corresponding docstring
        m.monitor_depo()

    Args:
        quartz (Instrument): The quartz oscillator instrument measuring thickness and rate.
        ohmmeter (Instrument): The resistance measurement instrument.
        mfc (Instrument): The mass flow controller instrument.
        film_name (str): A name can be given, that is added to the directory and filename.
        sample (Sample): A sample object can be given that is stored in the h5 file.

    Attributes:
        Should be listed here... Until then, use tabbing.
    """

    def __init__(self, quartz=None, ohmmeter=None, mfc=None, film_name='', sample=None):
        self.quartz = quartz
        self.ohmmeter = ohmmeter
        self.mfc = mfc

        self.film_name = film_name
        self._sample = sample

        self.comment = ''
        self.dirname = None

        self.x_set_obj = None
        self.y_set_obj = None

        self.progress_bar = False
        self.stopwatch = False

        self._show_ideal = False

        self._fit_resistance = False
        self._fit_every = 1
        self._fit_points = 5
        self._p0 = None

        self._reference_uid = None

        self._plot_comment = ""

        self.open_qviewkit = True
        self.qviewkit_singleInstance = False

        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'sputter_deposition'
        self._measurement_object.sample = self._sample

        self._qvk_process = True

        self._duration = 60.  # Duration of monitoring
        self._resolution = 2.  # How often values should be taken

        self._target_resistance = 1000.
        self._target_thickness = 20.
        self._target_marker = False

        self._depmon_queue = Queue()  # for communication with the depmon thread

    def set_duration(self, duration=60.):
        """
        Set the duration for the monitoring. Better too long than too short.

        Args:
            duration (float): Monitoring duration in seconds.
        """
        self._duration = duration

    def get_duration(self):
        """
        Get the duration for the monitoring.

        Returns:
            duration (float): Monitoring duration in seconds.
        """
        return self._duration

    def set_resolution(self, resolution=2.0):
        """
        Set the resolution for the monitoring.
        Every x second, the parameters will be read and stored.

        Args:
            resolution (float): Monitoring resolution in seconds.
        """
        self._resolution = resolution

    def get_resolution(self):
        """
        Get the resolution for the monitoring.
        Every x second, the parameters will be read and stored.

        Returns:
            resolution (float): Monitoring resolution in seconds.
        """
        return self._resolution

    def set_filmparameters(self, resistance=1000., thickness=20., target_marker=False):
        """
        Set the target parameters of the film to be deposited.
        They are used to calculate the ideal trend of the measured parameters and their deviation from the ideal.

        Args:
            resistance (float): The target sheet resistance in Ohm.
            thickness (float): The target thickness in nm.
            target_marker (bool): Display a cross marking the target resistance and thickness.
        """
        self._target_resistance = resistance
        self._target_thickness = thickness
        self._target_marker = target_marker

    def get_filmparameters(self):
        """
        Get the set film parameters.

        Returns:
            Tuple containing:
                [0]: target_resistance (float): The target sheet resistance in Ohm.
                [1]: target_thickness (float): The target thickness in nm.
        """
        return self._target_resistance, self._target_thickness

    def ideal_resistance(self, thickness):
        """
        Calculates the ideal resistance at a given thickness depending on the set film parameters.

        Args:
            thickness (float): Film thickness in nm for which the ideal resistance should be calculated.

        Returns:
            The ideal resistance as a float.
        """
        try:
            return 1. / (thickness * (1. / self._target_resistance) / self._target_thickness)
        except:
            return np.nan

    def ideal_trend(self):
        """
        Calculates a list of ideal resistances that is used as comparison in a view together with the real values.
        """
        t = np.linspace(1, self._target_thickness, 300)
        return [t, self.ideal_resistance(t)]

    def set_fit(self, fit=False, fit_every=1, fit_points=5, p0=None):
        """
        Settings for the fit used to estimate final resistance at target thickness and thickness at target resistance.

        Args:
            fit (bool): Turn on the fit.
            fit_every (int): Fit after every x measurement.
            fit_points (int): Fit to the last x points.
            p0 (list): Provide initial values for the curvefit.
        """
        self._fit_resistance = fit
        self._fit_every = fit_every
        self._fit_points = fit_points
        self._p0 = p0

    def set_reference_uid(self, uid=None):
        """
        Set a UID to be shown as a reference trace in the "resistance_thickness" view.

        Args:
            uid (str): The UID of the h5 file from which the resistance and thickness data is to be displayed.
        """
        try:
            ref = hdf.Data(qkit.fid[uid])
            self._ref_resistance = ref.data.resistance[:]
            self._ref_thickness = ref.data.thickness[:]
            self._reference_uid = uid
            ref.close()
            return self._reference_uid
        except:
            self._reference_uid = None
            return "Invalid UID"

    def set_resistance_4W(self, four_wire=False):
        """
        Sets 2 (default) or 4 wire measurement mode, if the device supports it.

        Args:
            four_wire = False
        """
        self.ohmmeter.set_measure_4W(four_wire)

    def get_rate(self, test_thickness):
        """
        Tests deposition rate.

        Args:
            test_thickness (in nm)
        """
        self.quartz.set_timer_zero()
        self.quartz.set_thickness_zero()
        time.sleep(3)
        while self.quartz.get_thickness() < test_thickness * 1e-2:
            time.sleep(1)
        mins, secs = self.quartz.get_time().split(':')
        t = 60 * float(mins) + float(secs)
        rate = test_thickness / t
        return rate, t

    def _theory(self, thickness, thickness_start, conductivity, cond_per_layer):
        """
        Calculate a prediction for the resistance at a certain film thickness,
        given the current thickness, current conductivity and added conductivity per layer.

        Args:
            thickness (float): Film thickness for which the resistance shall be predicted
            thickness_start (float): Film thickness at the start of the prediction.
            conductivity (float): Film conductivity at the start of the prediction.
            cond_per_layer (float): Added conductivity per layer for the remaining film growth.

        Returns:
            The projected film resistance (float).
        """
        t_added = thickness - thickness_start
        try:
            return 1. / (conductivity + t_added * cond_per_layer)
        except:
            return np.nan

    def _linear(self, x, a, b):
        """
        Helper function providing a linear fitting function.
        """
        return a * x + b

    def _fit_layer(self, t_points, R_points):
        """
        Fit the changing film resistance to extract the gained conductivity per layer.
        This is then used to project the final film resistance with the provided target film parameters.

        Args:
            t_points (numpy array): Thickness data from which the conductivity per layer is to be extracted.
            R_points (numpy array): Resistance data from which the conductivity per layer is to be extracted.

        Returns:
            Tuple containing:
                [0]: t_final (float): The predicted thickness at which the target resistance will be reached.
                [1]: R_final (float): The predicted resistance that will be reached at the target thickness.
                [2]: popt[0] (float): Conductivity per layer as obtained by linear fit.
        """
        # ToDo: Store all fits, Store time of fit to sync it to other datasets, revise theory.
        for i in R_points:
            if i == 0:
                return [np.nan, np.nan, np.nan]

        try:
            c_target = 1. / self._target_resistance
            c_points = 1. / R_points
            popt, _ = curve_fit(self._linear, t_points, c_points, p0=self._p0)
            R_final = self._theory(self._target_thickness, t_points[0], c_points[0], popt[0])
            t_final = ((c_target - c_points[-1]) / popt[0]) + t_points[-1]
            return [t_final, R_final, popt[0]]

        except:
            return [np.nan, np.nan, np.nan]

    def _reciprocal(self, thickness, cond_per_layer):
        """
        Function used by the curvefit routine.

        Args:
            thickness: Film thickness in nm.
            cond_per_layer: Film conductance per layer.

        Returns:
            The resistance for the given parameters as a float.
        """
        try:
            return 1. / (thickness * cond_per_layer)
        except:
            return np.nan

    def _fit_trend(self, t_points, R_points):
        """
        Do the fit used to estimate final resistance at target thickness and thickness at target resistance.

        Args:
            t_points: The last x points of film thickness in nm.
            R_points: The last x points of film resistivity in Ohm.

        Returns:
            List of [R_final, t_final] with
            R_final (float): Estimated final resistance at target thickness.
            t_final (float): Estimated thickness at target resistance.
        """
        for i in t_points:
            if i == 0:
                return [np.nan, np.nan]

        try:
            self._popt, _ = curve_fit(self._reciprocal, t_points, R_points, p0=self._p0)

            t_final = 1. / (self._target_resistance * self._popt[0])
            R_final = 1. / (self._target_thickness * self._popt[0])

            return [t_final, R_final]

        except:
            return [np.nan, np.nan]

    def _prepare_measurement_quartz(self):
        """
        All the relevant settings from the quartz are updated and called.
        """
        # self.quartz.get_all()
        # self.quartz.get_frequency() # Store it somewhere
        pass

    def _prepare_measurement_ohmmeter(self):
        """
        All the relevant settings from the ohmmeter are updated and called.
        """
        # self.ohmmeter.get_all()
        pass

    def _prepare_measurement_mfc(self):
        """
        All the relevant settings from the ohmmeter are updated and called.
        """
        # self.mfc.get_all()
        self.Ar_channel = self.mfc.predef_channels['Ar']
        self.ArO_channel = self.mfc.predef_channels['ArO']

        pass

    def _prepare_monitoring_file(self):
        """
        Creates the output .h5-file with distinct the required datasets and views.
        At this point all measurement parameters are known and put in the output file.
        """
        self._data_file = hdf.Data(name=self._file_name, mode='a')
        self._measurement_object.uuid = self._data_file._uuid
        self._measurement_object.hdf_relpath = self._data_file._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()

        self._measurement_object.save()
        self._mo = self._data_file.add_textlist('measurement')
        self._mo.append(self._measurement_object.get_JSON())

        # write logfile and instrument settings
        self._write_settings_dataset()
        self._log = waf.open_log_file(self._data_file.get_filepath())

        '''
        Time record
        '''
        self._data_time = self._data_file.add_value_vector('time', x=None, unit='s')

        '''
        Quartz datasets
        '''
        self._data_rate = self._data_file.add_value_vector('rate',
                                                           x=self._data_time,
                                                           unit='nm/s',
                                                           save_timestamp=False)
        self._data_thickness = self._data_file.add_value_vector('thickness',
                                                                x=self._data_time,
                                                                unit='nm',
                                                                save_timestamp=False)

        '''
        Ohmmeter datasets
        '''
        self._data_resistance = self._data_file.add_value_vector('resistance',
                                                                 x=self._data_time,
                                                                 unit='Ohm',
                                                                 save_timestamp=False)
        self._data_conductance = self._data_file.add_value_vector('conductance',
                                                                  x=self._data_time,
                                                                  unit='S',
                                                                  save_timestamp=False)
        self._data_deviation_abs = self._data_file.add_value_vector('deviation_absolute',
                                                                    x=self._data_time,
                                                                    unit='Ohm',
                                                                    save_timestamp=False)
        self._data_deviation_rel = self._data_file.add_value_vector('deviation_relative',
                                                                    x=self._data_time,
                                                                    unit='relative',
                                                                    save_timestamp=False)

        '''
        MFC datasets
        '''
        # FIXME: units?
        if self.mfc:
            self._data_pressure = self._data_file.add_value_vector('pressure',
                                                                   x=self._data_time,
                                                                   unit='ubar',
                                                                   save_timestamp=False)
            self._data_Ar_flow = self._data_file.add_value_vector('Ar_flow',
                                                                  x=self._data_time,
                                                                  unit='sccm',
                                                                  save_timestamp=False)
            self._data_ArO_flow = self._data_file.add_value_vector('ArO_flow',
                                                                   x=self._data_time,
                                                                   unit='sccm',
                                                                   save_timestamp=False)

        '''
        Calculate ideal trend and create record
        '''
        self._thickness_coord = self._data_file.add_coordinate('thickness_coord', unit='nm')
        self._thickness_coord.add(self.ideal_trend()[0])
        if self._show_ideal:
            self._data_ideal = self._data_file.add_value_vector('ideal_resistance',
                                                                x=self._thickness_coord,
                                                                unit='Ohm',
                                                                save_timestamp=False)
            self._data_ideal.append(self.ideal_trend()[1])

        '''
        Create target marker
        '''
        if self._target_marker:
            self._target_thickness_line = self._data_file.add_value_vector('target_thickness',
                                                                           x=None,
                                                                           unit='nm',
                                                                           save_timestamp=False)
            self._target_thickness_line.append([0.8 * self._target_thickness,
                                                self._target_thickness, self._target_thickness,
                                                self._target_thickness, self._target_thickness,
                                                1.2 * self._target_thickness])
            self._target_resistance_line = self._data_file.add_value_vector('target_resistance',
                                                                            x=None,
                                                                            unit='Ohm',
                                                                            save_timestamp=False)
            self._target_resistance_line.append([self._target_resistance, self._target_resistance,
                                                 1.2 * self._target_resistance, 0.8 * self._target_resistance,
                                                 self._target_resistance, self._target_resistance])
            self._target_conductance_line = self._data_file.add_value_vector('target_conductance',
                                                                             x=None,
                                                                             unit='S',
                                                                             save_timestamp=False)
            self._target_conductance_line.append([1. / self._target_resistance, 1. / self._target_resistance,
                                                  1. / 1.2 / self._target_resistance,
                                                  1. / 0.8 / self._target_resistance,
                                                  1. / self._target_resistance, 1. / self._target_resistance])

        '''
        Estimation datasets
        '''
        if self._fit_resistance:
            self._thickness_estimation = self._data_file.add_value_vector('thickness_estimation',
                                                                          x=None,
                                                                          unit='nm',
                                                                          save_timestamp=False)
            self._resistance_estimation = self._data_file.add_value_vector('resistance_estimation',
                                                                           x=None,
                                                                           unit='Ohm',
                                                                           save_timestamp=False)
            self._last_resistance_fit = self._data_file.add_value_vector('last_resistance_fit',
                                                                         x=None,
                                                                         unit='Ohm',
                                                                         save_timestamp=False)

        '''
        Reference dataset
        '''
        if not self._reference_uid == None:
            self._thickness_reference = self._data_file.add_value_vector('thickness_reference',
                                                                         x=None,
                                                                         unit='nm',
                                                                         save_timestamp=False)
            self._thickness_reference.append(self._ref_thickness)
            self._resistance_reference = self._data_file.add_value_vector('reference_' + self._reference_uid,
                                                                          x=None,
                                                                          unit='Ohm',
                                                                          save_timestamp=False)
            self._resistance_reference.append(self._ref_resistance)

        '''
        Create Views
        '''
        self._resist_view = self._data_file.add_view('resistance_thickness',
                                                     x=self._data_thickness,
                                                     y=self._data_resistance)
        if self._show_ideal:
            self._resist_view.add(x=self._thickness_coord, y=self._data_ideal)
        if self._target_marker:
            self._resist_view.add(x=self._target_thickness_line, y=self._target_resistance_line)
        if not self._reference_uid == None:
            self._resist_view.add(x=self._thickness_reference, y=self._resistance_reference)
        if self._fit_resistance:
            self._resist_view.add(x=self._thickness_coord, y=self._last_resistance_fit)

        self._conductance_view = self._data_file.add_view('conductance_thickness',
                                                          x=self._data_thickness,
                                                          y=self._data_conductance)
        if self._target_marker:
            self._conductance_view.add(x=self._target_thickness_line, y=self._target_conductance_line)

        self._deviation_abs_view = self._data_file.add_view('deviation_absolute',
                                                            x=self._data_thickness,
                                                            y=self._data_deviation_abs)

        self._deviation_rel_view = self._data_file.add_view('deviation_relative',
                                                            x=self._data_thickness,
                                                            y=self._data_deviation_rel)

        '''
        Create comment
        '''
        if self.comment:
            self._data_file.add_comment(self.comment)

        '''
        Open GUI
        '''
        if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
            self._qvk_process.terminate()  # terminate an old qviewkit instance

    def _write_settings_dataset(self):
        """
        Writes a dataset containing the settings of the measurement instruments.
        """
        self._settings = self._data_file.add_textlist('settings')
        settings = waf.get_instrument_settings(self._data_file.get_filepath())
        self._settings.append(settings)

    def _init_depmon(self):
        """
        Initializes the measurement, the required files and instruments.
        Should only be called by the main functions monitor_depo or start_depmon.
        """
        self._measurement_object.measurement_func = 'sputter_monitoring'
        self._measurement_object.x_axis = 'time'
        self._measurement_object.y_axis = ''
        self._measurement_object.z_axis = ''
        self._measurement_object.web_visible = True

        if not self.dirname:
            self.dirname = 'SPUTTER_monitoring'
        self._file_name = self.dirname.replace(' ', '').replace(',', '_')
        if self.film_name:
            self._file_name += '_' + self.film_name

        self.x_vec = np.arange(0, self._duration, self._resolution)

        self._prepare_measurement_quartz()
        self._prepare_measurement_ohmmeter()
        if self.mfc:
            self._prepare_measurement_mfc()

        self._prepare_monitoring_file()

    def _acquire(self, mon_data):
        """
        Acquires and stores current information from the instruments.
        Should only be called by main functions monitor_depo or start_depmon.
        """
        self._data_time.append(time.time() - mon_data.t0)

        # mon_data.resistance.append(self.ohmmeter.get_resistance())
        mon_data.resistance = np.append(mon_data.resistance, self.ohmmeter.get_resistance())
        if self.quartz:
            rate = self.quartz.get_rate(nm=True)
            # mon_data.thickness.append(self.quartz.get_thickness(nm=True))
            mon_data.thickness = np.append(mon_data.thickness, self.quartz.get_thickness(nm=True))
        if self.mfc:
            pressure = self.mfc.get_pressure()
            Ar_flow = self.mfc.get_flow(self.Ar_channel)
            ArO_flow = self.mfc.get_flow(self.ArO_channel)

        if mon_data.resistance[-1] > 1.e9:
            mon_data.resistance[-1] = np.nan
        self._data_resistance.append(mon_data.resistance[-1])
        if not mon_data.resistance[-1] == 0:
            self._data_conductance.append(1. / mon_data.resistance[-1])
        else:
            self._data_conductance.append(np.nan)
        if self.quartz:
            self._data_rate.append(rate)
            self._data_thickness.append(mon_data.thickness[-1])
        if self.mfc:
            self._data_pressure.append(pressure)
            self._data_Ar_flow.append(Ar_flow)
            self._data_ArO_flow.append(ArO_flow)
        if self.quartz:
            deviation_abs = mon_data.resistance[-1] - self.ideal_resistance(mon_data.thickness[-1])
            deviation_rel = deviation_abs / self._target_resistance

            self._data_deviation_abs.append(deviation_abs)
            self._data_deviation_rel.append(deviation_rel)

        if ((self._fit_resistance) and (mon_data.it % self._fit_every == 0) and (
                len(mon_data.resistance) >= self._fit_points)):
            """
            # OLD ROUTINE
            estimation = self._fit_trend(mon_data.thickness[-self._fit_points:None],
                                         mon_data.resistance[-self._fit_points:None])
            self._thickness_estimation.append(estimation[0])
            self._resistance_estimation.append(estimation[1])
            self._last_resistance_fit.append(self._reciprocal(self.ideal_trend()[0],self._popt[0]), reset=True)
            """
            estimation = self._fit_layer(mon_data.thickness[-self._fit_points:None],
                                         mon_data.resistance[-self._fit_points:None])
            self._thickness_estimation.append(estimation[0])
            self._resistance_estimation.append(estimation[1])
            self._last_resistance_fit.append(self._theory(self.ideal_trend()[0], mon_data.thickness[-1],
                                                          1. / mon_data.resistance[-1], estimation[2]),
                                             reset=True)

    def monitor_depo(self):
        """
        Main sputter deposition monitoring function.

        Consider using the threaded version by calling start_depmon().

        Records the film resistance, thickness and deposition rate live during the sputter process.
        Stores everything into a h5 file.
        Provides measures to estimate the resulting resistance of the final film and thereby
        facilitates live adjustments to the sputter parameters.

        Note:
            set_duration and set_resolution should be called before to set the monitoring length and time resolution.
            set_filmparameters should be called before to provide the actual target values.
        """
        
        self._init_depmon()

        if self.progress_bar:
            self._p = Progress_Bar(self._duration / self._resolution,
                                   'EVAP_timetrace ' + self.dirname,
                                   self._resolution)  # FIXME: Doesn't make much sense...
        
        if self.stopwatch:
            self._sw = Stopwatch()

        print('Monitoring deposition...')
        sys.stdout.flush()

        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=['views/resistance_thickness'])

        try:
            """
            Initialize the data lists.
            """

            class MonData(object):
                resistance = np.array([])
                thickness = np.array([])

            mon_data = MonData()

            """
            Main loop:
            """
            mon_data.t0 = time.time()  # note: Windows has limitations on time precision (about 16ms)
            for mon_data.it, _ in enumerate(self.x_vec):
                self._acquire(mon_data)

                if self.progress_bar:
                    self._p.iterate()
                
                if self.stopwatch:
                    self._sw.update(time.time() - mon_data.t0)

                # calculate the time when the next iteration should take place
                ti = mon_data.t0 + (float(mon_data.it) + 1) * self._resolution
                # wait until the total dt(iteration) has elapsed
                while time.time() < ti:
                    time.sleep(0.05)

        except KeyboardInterrupt:
            print('Monitoring interrupted by user.')

        except Exception as e:
            print(e)
            print(e.__doc__)
            print(e.message)

        finally:
            self._end_measurement()

    def _monitor_depo_bg(self):
        """
        Main sputter deposition monitoring function, threaded version

        Should only be called by main function start_depmon.

        Records the film resistance, thickness and deposition rate live during the sputter process.
        Stores everything into a h5 file.
        Provides measures to estimate the resulting resistance of the final film and thereby
        facilitates live adjustments to the sputter parameters.

        Note:
            set_duration and set_resolution should be called before to set the monitoring length and time resolution.
            set_filmparameters should be called before to provide the actual target values.
        """

        class StopPauseExeption(Exception):
            pass

        class StartPauseExeption(Exception):
            pass

        """
        A few boilerplate functions for the thread management:
        continue, stop, status, etc...
        """

        def stop():
            raise StopIteration

        def do_nothing():
            pass

        def status():
            self._depmon_queue.put(loop_status)

        def cont():
            raise StopPauseExeption

        def pause():
            raise StartPauseExeption

        loop_status = 0
        tasks = {}
        tasks[0] = stop
        tasks[1] = do_nothing
        tasks[2] = status
        tasks[3] = pause
        tasks[4] = cont

        """
        Initialize the data lists.
        """

        class MonData(object):
            resistance = np.array([])
            thickness = np.array([])

        mon_data = MonData()

        """
        Main loop:
        """
        mon_data.it = 0
        mon_data.t0 = time.time()  # note: Windows has limitations on time precision (about 16ms)?
        while True:
            self._acquire(mon_data)
                
            if self.stopwatch:
                self._sw.update(time.time() - mon_data.t0)

            # Handle external commands
            try:
                tasks.get(self._depmon_queue.get(False), do_nothing)()
                self._depmon_queue.task_done()
            except Empty:
                pass
            except StartPauseExeption:
                print('monitoring paused')
                while True:
                    try:
                        time.sleep(0.1)
                        tasks.get(self._depmon_queue.get(False), do_nothing)()
                        self._depmon_queue.task_done()
                    except StopPauseExeption:
                        print('monitoring restarted')
                        break
                    except Empty:
                        pass
            except StopPauseExeption:
                pass
            except StopIteration:
                break

            # calculate the time when the next iteration should take place
            mon_data.it += 1
            ti = mon_data.t0 + (float(mon_data.it)) * self._resolution
            loop_status = ti  # for now we simply return the runtime
            # wait until the total dt(iteration) has elapsed
            while time.time() < ti:
                time.sleep(0.05)

        self._end_measurement()

    def start_depmon(self):
        """
        Starts the main sputter deposition monitoring function background process (depmon).

        Records the film resistance, thickness and deposition rate live during the sputter process.
        Stores everything into a h5 file.
        Provides measures to estimate the resulting resistance of the final film and thereby
        facilitates live adjustments to the sputter parameters.

        Note:
            set_duration and set_resolution should be called before to set the monitoring length and time resolution.
            set_filmparameters should be called before to provide the actual target values.

        Hint:
            Check with 'list_depmon_threads()'
            Stop with 'stop_depmon()'
        """
        print('Monitoring deposition...')
        sys.stdout.flush()

        self._init_depmon()
        
        if self.stopwatch:
            self._sw = Stopwatch()

        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=['views/resistance_thickness'],
                                              refresh=1.)

        self._depmon = Thread(target=self._monitor_depo_bg, name="depmon-1")
        self._depmon.daemon = True
        self._depmon.start()

    def stop_depmon(self):
        """
        Stop the deposition monitoring thread.
        Can be executed several times if more than one monitoring thread is running.

        Hint:
            Check with 'list_depmon_threads()'
        """
        self._depmon_queue.put(0)

    def status_depmon(self):
        """
        Give a heartbeat of the deposition monitoring thread (in the moment just the runtime).
        """
        self._depmon_queue.put(2)
        print(self._depmon_queue.get())
        self._depmon_queue.task_done()

    def pause_depmon(self):
        print('Pause Monitoring deposition...')
        self._depmon_queue.put(3)

    def continue_depmon(self):
        print('Continue Monitoring deposition...')
        self._depmon_queue.put(4)

    def list_depmon_threads(self):
        """
        List all bg deposition monitoring threads.
        """
        for thread in threading.enumerate():
            if thread.getName()[:3] == "dep":
                print(thread)

    def _end_measurement(self):
        """
        The data file is closed and file path is printed.
        """
        print(self._data_file.get_filepath())
        # qviewkit.save_plots(self._data_file.get_filepath(),comment=self._plot_comment)
        # #old version where we have to wait for the plots
        t = threading.Thread(target=qviewkit.save_plots,
                             args=[self._data_file.get_filepath(), self._plot_comment])
        t.start()
        self._data_file.close_file()
        waf.close_log_file(self._log)
        self.dirname = None

    # FIXME: Use flow.sleep? Code there looks rather bulky and maybe not suited for high speed measurements


class SputterControl(object):
    pass



# qtLAB preliminary driver for RF frontend PCB
# filename: RFPCB.py
# Robert Gartmann <gartmann@kit.edu>, 2021

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import zerorpc
import types
import logging
import json
import warnings
from time import sleep
from enum import Enum

class RFPCB(Instrument):
    '''
    This is the pre-release python driver for the RF frontend PCB
    Usage:
        Initialise with
        <name> = instruments.create('<name>', "RFPCB", address='<TCP/IP>:<port>')
    '''

    def __init__(self, name, address, model = 'RFPCB', ifcal = False):
        '''
        Initializes the ZeroRPC Client to communicate with according Server (RPi).
        Input:
            name (string)    : name of the instrument
            address (string) : TCPIP Address including port e.g. "10.22.197.145:4242"
        '''

        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name)
        self._address = "tcp://" + address
        self._model = model
        # self._statusdict = {}
        self._rpc = zerorpc.Client()
        self._rpc.connect(self._address)
        
        self._frequency = 4e9
        self._power = -30
        
        self._ifcal = ifcal
        self._iffreq = 2
        self._rffreq = 8
        self._nominaltxatt = 30
        self._txifslopecomp = [[3.1, 1], [3.2, 3], [3.3, 5], [3.4, 7], [3.5, 8], [3.6, 9], [3.7, 10]] # PCB 1 ca
        #[[3.1, 1], [3.2, 3], [3.3, 5], [3.4, 7], [3.5, 8], [3.6, 9], [3.7, 10]] # PCB 1 ca
        #[[3.1, 1], [3.2, 3], [3.3, 5], [3.4, 7], [3.5, 9], [3.6, 10], [3.7, 10]] # Agressive PCB3 values
        #[[3.1, 1], [3.2, 2], [3.3, 3], [3.4, 4], [3.5, 5], [3.6, 6], [3.7, 7]] # Bare BB to IF 
        self._nominalrxatt = 30
        self._rxifslopecomp = [[2, 0], [2.3, 0], [2.8, 2], [3, 4], [3.2, 6], [3.4, 6.5], [3.6, 7], [3.8, 9], [3.9, 10]] # PCB 1 ca
        #[[2, 0], [2.3, 0], [2.8, 2], [3, 4], [3.2, 6], [3.4, 6.5], [3.6, 7], [3.8, 9], [3.9, 10]] # PCB 1 ca
        #[[2, 0], [2.3, 0], [2.8, 2], [3, 4], [3.2, 6], [3.4, 8], [3.6, 10], [3.8, 12], [4, 13]] # Agressive PCB3 values
        #[[2, 1], [2.3, 0], [2.8, 1], [3, 2], [3.2, 3], [3.4, 4], [3.6, 5], [3.8, 6], [4, 7]] # Bare IF to BB
        
        self.set_hf_attenuation(self._power)
        #self.set_frequency

        # Implement parameters
        self.add_parameter('frequency', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=4e9, maxval=12e9,
            units='Hz')
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-30, maxval=0,
            units='dBm')
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET)
        
        self.off()
        # Implement functions
        # self.get_all(True)

    # initialization related TODO: implement read function (drivers need to be done first)

    def set_rflo_powerdown(self, boolean):
        self._rpc.writeval("RFLO", "power_down", int(boolean))
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)

    def set_iflo_powerdown(self, boolean):
        self._rpc.writeval("ZFLO", "power_down", int(boolean))
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)

    def set_txpilot_powerdown(self, boolean):
        self._rpc.writeval("TXPilot", "power_down", int(boolean))
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_powerdown(self, boolean):
        self._rpc.writeval("RXPilot", "power_down", int(boolean))
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_rflo_enable_a(self, boolean):
        self._rpc.writeval("RFLO", "enable_a", int(boolean))
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)

    def set_iflo_enable_a(self, boolean):
        self._rpc.writeval("ZFLO", "enable_a", int(boolean))
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)

    def set_txpilot_enable_a(self, boolean):
        self._rpc.writeval("TXPilot", "enable_a", int(boolean))
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_enable_a(self, boolean):
        self._rpc.writeval("RXPilot", "enable_a", int(boolean))
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_rflo_enable_b(self, boolean):
        self._rpc.writeval("RFLO", "enable_b", int(boolean))
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)

    def set_iflo_enable_b(self, boolean):
        self._rpc.writeval("ZFLO", "enable_b", int(boolean))
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)

    def set_txpilot_enable_b(self, boolean):
        self._rpc.writeval("TXPilot", "enable_b", int(boolean))
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_enable_b(self, boolean):
        self._rpc.writeval("RXPilot", "enable_b", int(boolean))
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_rflo_frequency(self, value):
        self._rpc.writeval("RFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("RFLO", "recalibrate", 1)
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)
        self._rffreq = value/1e9

    def set_iflo_frequency(self, value):
        self._rpc.writeval("ZFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("ZFLO", "recalibrate", 1)
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)
        self._rpc.writeval("Demodulator", "lo_match_mhz", int(value/1e6))
        #self._rpc.writeval("Demodulator", "lo_match_mhz", 
        #    int(3000 if value > 3.9e9 else int(value/1e6)))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        self._iffreq = value/1e9
        #self.set_iflo_output_b_power(5 if value < 3.5e9 else 31)
        

    def set_txpilot_frequency(self, value):
        self._rpc.writeval("TXPilot", "vco_out_frequency", int(value))
        self._rpc.writeval("TXPilot", "recalibrate", 1)
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_frequency(self, value):
        self._rpc.writeval("RXPilot", "vco_out_frequency", int(value))
        self._rpc.writeval("RXPilot", "recalibrate", 1)
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_rflo_output_a_power(self, value):
        self._rpc.writeval("RFLO", "output_a_power", int(value))
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)

    def set_iflo_output_a_power(self, value):
        self._rpc.writeval("ZFLO", "output_a_power", int(value))
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)

    def set_txpilot_output_a_power(self, value):
        self._rpc.writeval("TXPilot", "output_a_power", int(value))
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_output_a_power(self, value):
        self._rpc.writeval("RXPilot", "output_a_power", int(value))
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_rflo_output_b_power(self, value):
        self._rpc.writeval("RFLO", "output_b_power", int(value))
        self._rpc.writeval("RFLO", "write_regs_to_chip", 1)

    def set_iflo_output_b_power(self, value):
        self._rpc.writeval("ZFLO", "output_b_power", int(value))
        self._rpc.writeval("ZFLO", "write_regs_to_chip", 1)

    def set_txpilot_output_b_power(self, value):
        self._rpc.writeval("TXPilot", "output_b_power", int(value))
        self._rpc.writeval("TXPilot", "write_regs_to_chip", 1)

    def set_rxpilot_output_b_power(self, value):
        self._rpc.writeval("RXPilot", "output_b_power", int(value))
        self._rpc.writeval("RXPilot", "write_regs_to_chip", 1)


    def set_linopt(self, value):
        if (value > 3000):
            print("Linopt would exceed safe limits!\n Defaulting to 3V55")
            value = 3000
        self._rpc.setlinopt(value)

    def set_hf_attenuation(self, value1): # ,value2
        #if (value2 == 0):
        self._nominaltxatt = abs(value1)
        self._rpc.SetHFAtt(value1, 0) # Could also set voltages manually
        

    def set_demod_attenuation(self, value):
        self._nominalrxatt = value
        self._rpc.writeval("Demodulator", "attenuation", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
#    def set_demod_band(self,value):
#        self._rpc.writeval("Demodulator", "band", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
#    def set_demod_dc_offset_i(self, value):
#        self._rpc.writeval("Demodulator", "dc_offset_i", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#    
#    def set_demod_dc_offset_q(self, value):
#        self._rpc.writeval("Demodulator", "dc_offset_q", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)

    def set_demod_gain(self, value):
        self._rpc.writeval("Demodulator", "gain", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    def set_demod_gain_error(self, value):
        self._rpc.writeval("Demodulator", "gain_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    #def set_demod_hd2i(self, value):
    #def set_demod_hd2q(self, value):
    #def set_demod_hd3i(self, value):
    #def set_demod_hd3q(self, value):
    #def set_demod_im2(self, value):
    #def set_demod_im3(self, value):
    #def set_demod_im3i(self, value):
    #def set_demod_im3q(self, value):
    #def set_demod_ip3(self, value):
    #def set_demod_lo_bias(self, value): 
#    def set_demod_lo_match_c1(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_c1", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#    
#    def set_demod_lo_match_c2(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_c2", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#        
#    def set_demod_lo_match_inductor(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_inductor", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    def set_demod_phase_error(self, value):
        self._rpc.writeval("Demodulator", "phase_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    def reinitialise(self):
        self._rpc.writeval("Demodulator", "init", 1)
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        self._rpc.writeval("ZFLO", "init", 1)
        self._rpc.writeval("RFLO", "init", 1)
        self._rpc.writeval("TXPilot", "init", 1)
        self._rpc.writeval("RXPilot", "init", 1)
        
    #def set_demod_rf_switch(self, value):

    def do_get_frequency(self, query=True):
        '''
        Get frequency of device
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            microwave frequency (Hz)
        '''
        # TODO: read from LOs to calculate band
        return self._frequency
        
    def do_set_frequency(self, frequency, offset = 100e6, method="mixsweep", rfmult=1):
        '''
        Set frequency of device.
        Default BB frequency is 100MHz.
        Default method is sweeping the RF LO.
        Input:
            freq (float) : Frequency in Hz
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting frequency to %s Hz' % (frequency*1.0e9))
        
        iflo = 4e9 #Filtered 1,8 GHz to 5 GHz
        rflo = 8e9 #Filtered 3,8 GHz to 8,4 GHz
        
        if(method == "ifsweep"):
            # RF LO stationary, IF variable, USB
            # This can reduce spurs from RF LO harmonics.
            # IF drops by 10dB towards 4 GHz though.
            if(frequency <= 6e9 or frequency >= 10e9):
                rflo = 8e9
            elif(frequency > 6e9 and frequency <= 8e9):
                rflo = 4e9
            elif(frequency > 8e9 and frequency < 10e9):
                rflo = 6e9
            else:
                print("No valid LO configuration found!")
            iflo = abs(frequency - rflo) + offset
        elif(method == "ifsweep_extend"):
            # IF level drops even further for 5 GHz.
            if(frequency <= 6.5e9):
                rflo = 8.5e9
            elif(frequency > 6.5e9 and frequency < 9e9):
                rflo = 4e9
            elif(frequency >= 9e9 and frequency < 12e9):
                rflo = 7e9
            else:
                print("No valid LO configuration found!")
            iflo = abs(frequency - rflo) + offset
        elif(method == "rfsweep"):
            # IF LO stationary 2.3G => highest amplitude, RF variable
            # Low IF causes in band spurs (DSB)!
            iflo = 2.3e9 - offset # 2.3G with coupler for optimal level
            if(frequency <= 6.1e9):
                rflo = frequency + iflo 
            elif(frequency > 6.1e9 and frequency < 10.7e9):
                rflo = frequency - iflo
            elif(frequency >= 10.7):
                rflo = 8.4e9
                iflo = frequency - rflo 
            else:
                print("No valid LO configuration:")
            iflo += offset
        elif(method == "rfsweep_extend"):
            iflo = 4.5e9 - offset # RFLO Filter (<8G) & output filter (<12G) swapped
            rflo = frequency - iflo 
            if (rflo < 3.8e9):
                rflo = frequency + iflo
            iflo += offset
        elif(method == "mixsweep"):
            if(frequency <= 5e9):
                iflo = rflo - frequency + offset
            elif(frequency <= 9e9):
                iflo = 3e9  #### EDIT TODO: ALARM, soll 3G, 2G ist nur Test
                rflo = frequency + iflo - offset
            elif(frequency > 9e9 and frequency < 10e9):
                rflo = 12e9
                iflo = rflo - frequency - offset
            else:
                print("No valid LO configuration found!")
        else:
            print("Please select linear combination method!")
        
        self._iffreq = iflo/1e9
        self._rffreq = rflo/1e9
        
        if self._ifcal:
            txindex = 0
            rxindex = 0
            for i in range(0, len(self._txifslopecomp)):
                if (self._iffreq > self._txifslopecomp[i][0]):
                    txindex = i
            for i in range(0, len(self._rxifslopecomp)):
                if (self._iffreq > self._rxifslopecomp[i][0]):
                    rxindex = i
            #print("Indices: {:d} \t {:d}".format(txindex, rxindex))
            newtxatt = self._nominaltxatt - self._txifslopecomp[txindex][1]
            newrxatt = self._nominalrxatt - self._rxifslopecomp[rxindex][1]
            #print("Att: {:d}/{:d} \t {:d}/{:d}".format(
            #    newtxatt, self._nominaltxatt, newrxatt, self._nominalrxatt))
            if newtxatt < 0:
                newtxatt = 0
                print("TX slope cannot be compensated!")
            if newrxatt < 0:
                newrxatt = 0
                print("RX slope cannot be compensated!")
            self._rpc.SetHFAtt(newtxatt, 0)
            self._rpc.writeval("Demodulator", "attenuation", int(newrxatt))
            self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        self._frequency = frequency
        self.set_iflo_frequency(iflo)
        self.set_rflo_frequency(rflo/rfmult)

    def do_get_power(self, query=True):
        '''
        Get output power of device
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            microwave power (dBm)
        '''
        #TODO: implement read
        return self._power

    def do_set_power(self, power = 0):
        '''
        Sets 0 to 30 dB RF attenuation.
        Output level depends on modulator input or pilot power
        (usually around -10 to 0 dBm max.)
        Input:
            power (float) : "Power" in dBm
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting power to %s dBm' % power)
        self._power = power
        self.set_hf_attenuation(abs(power))

    def do_get_status(self, query=True):
        '''
        Get status of output channel
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            True (on) or False (off)
        '''
        #self.get_all(query)
        # TODO: implement read
        return self._status

    def do_set_status(self, status):
        '''
        Set status of output (Bool)
        Input:
            status (bool): enable LOs or disable everything
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting status to "%s"' % status)
        if status:
            self.set_rflo_enable_a(1)
            self.set_rflo_enable_b(1)
            self.set_rflo_output_a_power(0)
            self.set_rflo_output_b_power(0)

            self.set_iflo_enable_a(1)
            self.set_iflo_enable_b(1)
            self.set_iflo_output_a_power(0)
            self.set_iflo_output_b_power(31)
            
            self.set_rflo_powerdown(0)
            self.set_iflo_powerdown(0)
        else:
            self.set_rflo_powerdown(1)
            self.set_iflo_powerdown(1)
            self.set_txpilot_powerdown(1)
            self.set_rxpilot_powerdown(1)
        self._status = bool(status)


    #shortcuts
    def off(self):
        '''Turn everything Off'''
        self.set_status(0)

    def on(self):
        '''Turn LOs On'''
        self.set_status(1)



# OxfordInstruments_Kelvinox_IGH.py class, to perform the communication between the Wrapper and the device
# Guenevere Prawiroatmodjo <guen@vvtp.tudelft.nl>, 2009
# Pieter de Groot <pieterdegroot@gmail.com>, 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import time
import numpy as np
import serial


class OxfordInstruments_Kelvinox_IGH_serial(object):

    '''
    This is the python driver for the Oxford Instruments Kelvinox IGH Dilution Refrigerator and
    Intelligent Dilution Refrigerator Power Supply (IDR PS).

    Note: Since the ISOBUS allows for several instruments to be managed in parallel, the command
    which is sent to the device starts with '@n', where n is the ISOBUS instrument number.
    '''

#TODO: auto update script
#TODO: get doesn't always update the wrapper! (e.g. when input is an int and output is a string)

    def __init__(self, device='/dev/ttyUSB0'): 

        '''
        Initializes the Oxford Instruments Kelvinox IGH Dilution Refrigerator.

        Input:
            optional device (string)     :  '/dev/ttyUSB1'

        Output:
            None
        '''

        self.ser = serial.Serial(device, 9600, timeout=1)

    def remote_cmd(self, cmd):
		
        cmd += '\r\n'
        self.ser.write(cmd)
        time.sleep(0.3)
        rem_char = self.ser.inWaiting()
        value = self.ser.read(rem_char)
        value = value.replace('\xffR','',)
        value = value.replace('+','')
        value = value.replace('\r','')
	
        return value

###### TODO

    def identify(self):
        '''
        Identify the device

        Input:
            None
        Output:
            None
        '''
        #return self._execute('V')

    def remote(self):
        '''
        Set control to remote & unlocked

        Input:
            None
        Output:
            None
        '''
        # Set control to local & locked
        #self.set_remote_status(3)

    def local(self):
        '''
        Set control to local & locked

        Input:
            None
        Output:
            None
        '''
        # Set control to local & locked
        #self.set_remote_status(0)

    def set_remote_status(self, mode):
        '''
        Set remote control status.

        Input:
            mode(int) :
            0 : "Local and locked",
            1 : "Remote and locked",
            2 : "Local and unlocked",
            3 : "Remote and unlocked",

        Output:
            None
        '''
        
		#status = {
        #0 : "Local and locked",
        #1 : "Remote and locked",
        #2 : "Local and unlocked",
        #3 : "Remote and unlocked",
        #}
        #self._execute('C%s' % mode)

############

    def get_1K_pot_temp(self):

        '''
        Get 1K Pot Temperature from device.
        Input:
            None

        Output:
            result (float) : 1K Pot Temperature in mK
        '''
        result = self.remote_cmd("R2")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0

    def get_mix_chamber_temp(self):

        '''
        Get Mix Chamber Temperature
        Input:
            None
        Output:
            result (float) : Mix Chamber Temperature in mK
        '''
        result = self.remote_cmd("R3")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0

    def get_G1(self):
        '''
        Get Pressure G1
        Input:
            None

        Output:
            result (float) : G1 Pressure in mBar
        '''
        result = self.remote_cmd("R14")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0

    def get_G2(self):
        '''
        Get Pressure G2
        Input:
            None

        Output:
            result (float) : G2 Pressure in mBar
        '''
        result = self.remote_cmd("R15")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0

    def get_G3(self):
        '''
        Get Pressure G3
        Input:
            None

        Output:
            result (float) : G3 Pressure in mBar
        '''
        result = self.remote_cmd("R16")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0

    def get_P1(self):
        '''
        Get Pressure P1
        Input:
            Non

        Output:
            result (float) : P1 Pressure in mBar
        '''
        result = self.remote_cmd("R20")
	
	try:
	    return float(result)
	except ValueError:
	    return 0.0


















import socket
import select

class TCPClient():

    def __init__(self, host, port, packet_len=False):
        self._packet_len = packet_len
        self._socket = None
        self._connect(host, port)

    def _connect(self, host, port):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((host, port))

    def send(self, data):
        self._socket.send(data)

    def recv(self, maxsize, timeout=1):
        rlist = [self._socket.fileno()]
        elist = []
        lists = select.select(rlist, elist, elist, timeout)
        if len(lists[0]) == 0:
            return None

        data = self._socket.recv(maxsize)
        return data



"""Module to provide a high-level possibility to arange pulses for an experiment."""
from enum import Enum
import numpy as np
import inspect
from typing import Dict, Set, List, Union, Callable, Any, Tuple
import logging


plot_enable = False
try:
    import matplotlib.pyplot as plt

    plot_enable = True
except ImportError:
    pass


class Shape(np.vectorize):
    """
    A vectorized function describing a possible shape
    defined on the standardized interval [0,1).
    """

    def __init__(
        self, name: str, func: Callable[[float], float], *args: Any, **kwargs: Any
    ):
        self.name = name
        super(Shape, self).__init__(func, *args, **kwargs)

    def __mul__(self, other):
        return Shape(self.name, lambda x: self.pyfunc(x) * other.pyfunc(x))


class ShapeLibClass(object):
    """
    Object containing pre-defined pulse shapes.
    Currently implemented: rect, gauss
    """

    def __init__(self):
        self.zero = Shape("", lambda x: 0)
        self.rect = Shape("rect", lambda x: np.where(x >= 0 and x < 1, 1, 0))
        self.gauss = (
            Shape("gauss", lambda x: np.exp(-0.5 * np.power((x - 0.5) / 0.166, 2.0)))
            * self.rect
        )
        self.ramp = Shape("ramp", lambda x: x) * self.rect
        self.sqrfct = Shape("sqrfct", lambda x: x ** 2) * self.rect


# Make ShapeLib a singleton:
ShapeLib = ShapeLibClass()


class PulseType(Enum):
    """Type of Pulse object"""

    Pulse = 1
    Wait = 2
    Readout = 3


class ParametrizedValue:
    """
    class for calculating the value of a lambda expression
    value: Type lambda or float
    """

    Type = Union[float, Callable[..., float]]
    """The type of the parametrized values as given by the user."""

    def __init__(self, value: Type, name: str = ""):
        if isinstance(value, int):
            value = float(value)
        self.name = name
        self.value = value
        self._variables: Set[str] = set()
        if callable(value):
            self._variables.update(inspect.getfullargspec(self.value).args)

    def __call__(self, **kwargs: Any) -> float:
        if self.is_parametrized:
            useful_variables: Dict = {}
            for key, value in kwargs.items():
                if key in self.variables:
                    useful_variables[key] = value
            return self.value(**useful_variables)
        else:
            return self.value

    def crop_string(self, function: str) -> str:
        function = function[function.find(":") + 1 :]
        output = function[: function.find(",")]
        output = output.strip(" )\\n']")
        return output

    def __str__(self) -> str:
        if isinstance(self.value, float):
            return str(self.value)
        if self.value.__name__ != "<lambda>":
            function_line = inspect.getsourcelines(self.value)
            function_line_start = str(function_line).find(":") + 1
            return str(function_line)[6 : function_line_start - 1].strip()
        else:
            function = str(inspect.getsourcelines(self.value)[0])
            counter = str(function).count(":")
            if counter == 1:
                return self.crop_string(function)
            elif counter == 2:
                if self.name == "length":
                    if "length" in function:
                        function_start = function.find("length")
                        function = function[function_start:]
                        return self.crop_string(function)
                    else:
                        return self.crop_string(function)
                elif self.name == "amplitude":
                    if "amplitude" in function:
                        function_start = function.find("amplitude")
                        function = function[function_start:]
                        return self.crop_string(function)
                    else:
                        cutting_point = function.find("lambda")
                        function = function[cutting_point + 6 :]
                        function_start = function.find("lambda")
                        function = function[function_start:]
                        return self.crop_string(function)
                else:
                    raise Exception(
                        "unknown Parameter only length and amplitude are supportet"
                    )
            else:
                raise Exception(
                    "too many lambda functions. Only length and amplitude can be lambda functions"
                )

    @property
    def variables(self) -> Set[str]:
        return self._variables

    @property
    def is_parametrized(self) -> bool:
        return callable(self.value)


class Pulse(object):
    """
    Class to describe a single pulse.
    """

    def __init__(
        self,
        length: ParametrizedValue.Type,
        shape: Shape = ShapeLib.rect,
        name: str = None,
        amplitude: ParametrizedValue.Type = 1.0,
        phase: float = 0.0,
        iq_frequency: float = 0.0,
        iq_dc_offset: float = 0.0,
        iq_angle: float = 90.0,
        q_rel: float = 1.0,
        ptype=PulseType.Pulse,
    ):
        """
        Inits a pulse with:
            length:       length of the pulse. This can also be a (lambda) function for variable pulse lengths.
            shape:        pulse shape (i.e. rect, gauss, ...)
            name:         name you want to give your pulse (i.e. pi-pulse, ...)
            amplitude:    relative amplitude of your pulse. This can also be a (lambda) function for variable pulse amplitudes.
            phase:        phase of the pulse in deg. (i.e. 90 for pulse around y-axis of the bloch sphere)
            iq_frequency: IQ-frequency of your pulse for heterodyne mixing (if 0 homodyne mixing is employed)
            iq_dc_offset: complex dc offset for calibrating the IQ-mixer (real part for dc offset of I, imaginary part is dc offset of Q)
            iq_angle:     angle between I and Q in the complex plane (default is 90 deg)
            q_rel:        relative amplitude of Q in respect to I. This is needed for mixer calibration. If q_rel > 1 make sure you are still within the limits of your device.
            type:         The type of the created pulse (from enum PulseType: can be Pulse, Wait or Readout)
        """

        self.shape = shape
        self.name = name or ""
        self.amplitude = ParametrizedValue(amplitude, name="amplitude")
        self.length = ParametrizedValue(length, name="length")
        self.phase = phase
        self.iq_frequency = iq_frequency
        self.iq_dc_offset = iq_dc_offset
        self.iq_angle = iq_angle
        self.q_rel = q_rel
        self.type = ptype

    def __call__(
        self,
        samplerate: float,
        heterodyne: bool = False,
        start_phase: float = 0,
        **variables: Any
    ) -> np.ndarray:
        """
        Returns the pulse envelope for a given frequency and, if length is a function, with given variables as kwargs

        Args:
             samplerate: sample rate for calculating the envelope
             heterodyne: Bool if True returns a complex envelope (defaults to False)
             start_phase: the global phase at which the pulse should start (in rad, defaults to 0)
            **variables: the variables for the length/amplitude function, if any

        Returns:
            envelope of the pulse as numpy array.
        """
        length = self.length(**variables)
        amplitude = self.amplitude(**variables)
        timestep = 1.0 / samplerate

        if length < timestep / 2.0:
            if length != 0:
                logging.warning(
                    "The pulse '{:}' is shorter than {:.2f} ns and thus is omitted.".format(
                        self.name, timestep / 2.0 * 1e9
                    )
                )

            return np.zeros(0)

        time = np.arange(0, length, timestep)
        time_fractions = time / length

        if time_fractions[-1] >= 1.0:
            # This can happen due to float rounding error -> cut it away
            # (the shapes are only defined on [0,1) where 1 is not included)
            time = time[:-1]
            time_fractions = time_fractions[:-1]

        envelope = amplitude * self.shape(time_fractions)
        if not heterodyne or envelope.size == 0 or self.iq_frequency == 0:
            return envelope
        # Empty envelope needs no IQ modulation and
        # for homodyne mixing the envelope is real
        else:
            envelope = envelope * np.exp(
                1.0j
                * (
                    start_phase
                    - np.pi / 180 * self.phase
                    + 2 * np.pi * self.iq_frequency * time
                )
            )

        # account for mixer calibration i.e. dc offset and phase != 90deg between I and Q
        if self.iq_angle != 90 or self.q_rel != 1.0:
            envelope_i = np.real(envelope)
            envelope_q = self.q_rel * np.imag(
                envelope * np.exp(1.0j * np.pi / 180 * (90 - self.iq_angle))
            )
            envelope = envelope_i + 1.0j * envelope_q
        envelope[envelope != 0] += self.iq_dc_offset

        return envelope

    @property
    def variable_names(self) -> Set[str]:
        """A set with the names of all variables necessary to calculate the pulse length and amplitude."""
        return self.length.variables.union(self.amplitude.variables)

    @property
    def is_parametrized(self) -> bool:
        """False if only fixed Values for length/amplitude"""
        return self.length.is_parametrized or self.amplitude.is_parametrized


class PulseSequence(object):
    """
    Class for aranging pulses for a time-domain experiment.
    Sequence objects are callable, returning the sequence envelope for a given time step.
    Add wait as variable times in the experiment.
    Add readout to synchornize different channels in more sophisticated experiments.

    Attributes:
        add:         adds a given pulse to the experiment
        add_wait:    adds a wait time
        add_readout: adds the readout to the experiment
        plot:        plots schematic of the sequence
        get_pulses:  returns list of currently added pulses and their properties.
    """

    def __init__(
        self, sample: Any = None, samplerate: float = None, dc_corr: float = 0
    ):
        """
        Inits PulseSequence with sample and samplerate:
            sample:     Sample object
            samplerate: Samplerate of your device
                        This should already be specified in your sample object as sample.clock
            dc_corr:    DC Voltage bias of the AWG for idling times (Real p)complex dc offset for calibrating the IQ-mixer during idling times.
                        The real part encodes the dc offset of I, the imaginary part is the dc offset of Q.
                        This correction is added to the dc offset during the pulse (i.e. of the pulse object).
        """
        self._sequence: List[List[Pulse]] = []
        self._next_pulse_is_parallel = False
        self._pulses: Dict[str, Pulse] = {}
        self._variables: Set[str] = set()
        self._sample = sample
        self.dc_corr: float = dc_corr
        try:
            self.samplerate = self._sample.clock
        except AttributeError:
            self.samplerate = samplerate

        self._color_palette: List[str] = [
            "C0",
            "C1",
            "C2",
            "C3",
            "C4",
            "C5",
            "C6",
            "C8",
            "C9",
            "r",
            "g",
            "b",
            "y",
            "k",
            "m",
        ]
        self._pulse_cols: Dict[PulseType, str] = {
            PulseType.Readout: "C7",
            PulseType.Wait: "w",
        }

    def __call__(
        self,
        IQ_mixing: bool = False,
        include_readout: bool = False,
        samplerate: float = None,
        **variables: Any
    ) -> Union[Tuple[np.ndarray, int], None]:
        """
        Returns the envelope of the whole pulse sequence for the input time.
        Also returns the index where the readout pulse starts.
        If no readout tone is found it is assumed to be at the end of the sequence.

        Args:
            IQ_mixing:   returns complex valued sequence if IQ_mixing is True (real part encodes I, imaginary part encodes Q)
            include_readout:   If the readout pulse should be included in the resulting waveform
            **variables:    function arguments for time dependent pulse lengths/wait times. Parameter names need to match time function parameters.


        Returns:
            waveform:      numpy array of the squence envelope, if IQ_mixing is True real part is I, imaginary part is Q
            readout_index: index of the readout tone
        """

        if self._variables and self._variables != set(variables.keys()):
            logging.error(
                "Given function arguments do not match with required ones. "
                + "The following keyword arguments are required: {}.".format(
                    ", ".join(self._variables)
                )
            )
            return None

        samplerate = samplerate or self.samplerate
        if not samplerate:
            logging.error("Sequence call requires samplerate.")
            return None

        # build the waveform of this sequence
        full_waveform = np.zeros(0, dtype=np.complex128)
        timestep = 1.0 / samplerate  # minimum time step
        readout_index = 0  # index of the readout in the waveform of the whole sequence
        position_of_next_slice = 0  # index where the next time slice will start
        for time_slice in self._sequence:
            # holds all waveforms that start at the same time (within the same slice)
            wfm_slice = np.zeros(0, dtype=np.complex128)
            # tracks the length of the last waveform in the slice as the next slice will start after that
            last_wfm_length = 0
            for pulse in time_slice:
                # create waveform array of the current pulse
                # adjust global phase relative to the beginning of the sequence
                # (startphase is only relevant when IQ_mixing is True)
                startphase = (
                    2.0 * np.pi * pulse.iq_frequency * position_of_next_slice * timestep
                )  # zero for homodyne mixing
                wfm = pulse(
                    samplerate,
                    start_phase=startphase,
                    heterodyne=IQ_mixing,
                    **variables
                )

                # if (pulse_dict["pulse"].type == PulseType.Readout) and (i == len(self._sequence) - 1):
                #     # if readout is last, omit the wfm (apart from a single digit)
                #     length = timestep

                # Store index if this pulse is a readout pulse (will have the last one at the end)
                if pulse.type == PulseType.Readout:
                    readout_index = position_of_next_slice
                    if not include_readout:
                        # Readout pulses are not taken into account here
                        wfm[:] = 0

                # Store the size of the last waveform in a slice
                # This waveform has skip=False and thus the next slice will start when this pulse is finished
                # even if other pulses of the current slice are longer
                last_wfm_length = len(wfm)

                # Enlarge waveforms to be the same size
                if len(wfm_slice) < len(wfm):
                    wfm_slice.resize(len(wfm))
                else:
                    wfm.resize(len(wfm_slice))
                # Add pulse to waveform of current time slice
                wfm_slice += wfm

            # Resize waveform to be capable of holding current wfm_slice
            new_waveform_length: float = max(
                len(full_waveform), position_of_next_slice + len(wfm_slice)
            )
            full_waveform.resize(new_waveform_length)
            # Add current time slice to global waveform
            full_waveform[
                position_of_next_slice : (position_of_next_slice + len(wfm_slice))
            ] += wfm_slice

            # Update position for next slice (the last waveform has no skip and thus decides the time)
            position_of_next_slice += last_wfm_length

        full_waveform += self.dc_corr

        # make sure first and last point of the waveform go to 0
        full_waveform = np.append(0, full_waveform)
        full_waveform = np.append(full_waveform, 0)

        if not any(np.iscomplex(full_waveform)):
            # No complex information in there, so just return the real part
            full_waveform = np.real(full_waveform)

        return full_waveform, readout_index + 1  # +1 due to leading 0

    def add(self, pulse: Pulse, skip: bool = False):
        """
        Append a pulse to the sequence.

        Args:
            pulse: pulse object
            skip:  if True the next pulse in the sequence will not wait until this pulse is finished (i.e. they happen at the same time)
        """
        # Check if pulse name is valid and unique
        if pulse.name is None or not isinstance(pulse.name, str):
            logging.error("The pulse name has to be a string and must not be None.")
            return self
        elif pulse.name in self._pulses and not self._pulses[pulse.name] is pulse:
            logging.error(
                "Another pulse with the same name ({name}) is already present in the sequence!".format(
                    name=pulse.name
                )
            )
            return self

        # Add the pulse to the pulse dictionary if it is not yet present
        if not pulse.name in self._pulses:
            self._pulses[pulse.name] = pulse

            # Keep track of all variable names: Add them to a set of unique variable names
            self._variables.update(pulse.variable_names)

        if not self._next_pulse_is_parallel:
            # Add empty list for next pulse
            self._sequence.append([])
        # Add pulse to last sequence slice
        self._sequence[-1].append(pulse)

        # If skip is true the next pulse will be scheduled at the same time
        self._next_pulse_is_parallel = skip

        return self

    def add_wait(self, time: ParametrizedValue.Type, name: str = None):
        """
        Add a wait time to the sequence.
        Use a (lambda) function for variable wait times.

        Parameters
        ----------
        self : PulseSequence
            the PulseSequence
        time : ParametrizedValue.Type
            the waiting period as float or function
        name : str, optional
            A special name can be passed for this wait block (by default, wait[#] will be used)
        skip : bool, optional
            if this wait should be parallel to the next item in the sequece, by default False

        Returns
        -------
        PulseSequence
            the PulsSequence with an added waiting period
        """

        def compose_name(index: int) -> str:
            return "wait[{}]".format(index)

        if name is None:
            # Find a unused name for the next wait "pulse"
            wait_index = 0
            while compose_name(wait_index) in self._pulses:
                wait_index += 1
            name = compose_name(wait_index)

        wait_pulse = Pulse(time, shape=ShapeLib.zero, name=name, ptype=PulseType.Wait)
        return self.add(wait_pulse)

    def add_readout(self, pulse: Pulse = None, skip: bool = False):
        """
        Add a readout pulse to the sequence.

        Parameters
        ----------
        skip : bool, optional
            If True the next pulse will follow at the same time as the readout, by default False.
        pulse : Pulse, optional
            pulse: A user-defined readout pulse can be specified if necessary, by default rectangular shaped Pulse with lengtht as specified in sample.

        Returns
        -------
        PulseSequence
            The PulseSequence with an added Readout
        """

        def compose_name(index: int) -> str:
            return "readout[{}]".format(index)

        if pulse is None:
            # Find a unused name for the next readout pulse
            readout_index = 0
            while compose_name(readout_index) in self._pulses:
                readout_index += 1
            name = compose_name(readout_index)

            # Try to determine useful readout tone length
            try:
                readout_length = self._sample.rec_pulselength
            except AttributeError:
                try:
                    readout_length = self._sample.readout_tone_length
                except AttributeError:
                    readout_length = 0.0

            readout_pulse = Pulse(readout_length, name=name, ptype=PulseType.Readout)
        else:
            # If a special pulse is needed, user can add it
            readout_pulse = pulse

            if readout_pulse.type != PulseType.Readout:
                readout_pulse.type = PulseType.Readout
                logging.warning(
                    "The type of the added pulse has to be Readout and was changed accordingly."
                )

        return self.add(readout_pulse, skip)

    @property
    def variable_names(self) -> Set[str]:
        """A set with the names of all variables present in this sequence."""
        return self._variables

    def get_pulses(self) -> List[Dict[str, Any]]:
        """
        Returns a list of all pulses and their properties.
        The properties of each pulse are stored in a dictionary with keys: name, shape, length, skip value
        """
        dict_list: List[Dict[str, Any]] = []
        for time_slice in self._sequence:
            for i, pulse in enumerate(time_slice):
                # This is more for legacy reasons
                dict_list.append(
                    {
                        "name": pulse.name,
                        "shape": pulse.shape.name,
                        "length": pulse.length,
                        "iq_frequency": pulse.iq_frequency,
                        "phase": pulse.phase,
                        "skip": i != len(time_slice) - 1,
                    }
                )
        return dict_list

    @property
    def pulses(self) -> Dict[str, Pulse]:
        """A dictionary containing pulse name as key and pulse as value"""
        return self._pulses

    @property
    def sequence(self) -> List[List[Pulse]]:
        """A List of Lists containing pulses. If pulses are paralles they will be in the same sublist"""
        return self._sequence

    def plot(self):
        """
        Plot a schematic of the stored pulses.
        """
        if not plot_enable:
            raise ImportError("matplotlib not found.")

        fig, ax = plt.subplots()
        amp = 1
        ampmax = 1
        remaining_colors = self._color_palette[:]
        pulse_colors: Dict[str, str] = {}

        for i, time_slice in enumerate(self._sequence):
            for amp, pulse in enumerate(reversed(time_slice)):
                ampmax = max(ampmax, amp + 1)

                # Generate displayed text
                text = "{name}\n{shape}\n{time}".format(
                    name=pulse.name,
                    shape=pulse.shape.name,
                    time=(pulse.length if pulse.type != PulseType.Readout else ""),
                )

                if pulse.iq_frequency not in [0, None]:
                    text += "\n\n f_iq = {:.0f} MHz".format(pulse.iq_frequency / 1e6)
                    if pulse.phase != 0:
                        text += "\n phase = {:.0f} deg".format(pulse.phase)

                # Make sure pulse colors are unique
                if not remaining_colors:
                    remaining_colors = self._color_palette[:]
                    print("All colors already in use...\n Resetting color palette.")
                if pulse.type in self._pulse_cols.keys():
                    # Pulse type is special and has predefined color
                    col = self._pulse_cols[pulse.type]
                elif pulse.name in pulse_colors.keys():
                    # Pulse was in sequence before, take same color again
                    col = pulse_colors[pulse.name]
                else:
                    col = remaining_colors[0]
                    pulse_colors[pulse.name] = col
                    remaining_colors = remaining_colors[1:]

                ax.fill(
                    [i, i, i + 1, i + 1, i],
                    [amp, amp + 1, amp + 1, amp, amp],
                    color=col,
                    alpha=0.3,
                )
                ax.text(
                    i + 0.5,
                    amp + 0.5,
                    text,
                    horizontalalignment="center",
                    verticalalignment="center",
                )

        # make sure plot looks nice and fits on the screen (max number of pulses before scaling down is 9)
        size = 2.0 * min(1.0, 9.0 / len(self._sequence))
        fig.set_figheight(size * ampmax)
        fig.set_figwidth(size * (len(self._sequence) - 1) + 2.0)
        ax.set_xlabel("pulse number")
        ax.set_xticks(np.arange(len(self._sequence)))
        plt.xlim(
            -0.05,
        )
        # hide y ticks
        ax.set_yticks([])
        # hide top and right spines
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        return



#!/usr/bin/env python

# SR_830 class, derived from Picowatt_AVS47 class by Hannes Rotzinger.
# Jochen Zimmer, June 2012 -- ALPHA release version 0.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


QTLAB = True

if QTLAB:
    from qkit.core.instrument_base import Instrument
else:
    class Instrument(object):
        FLAG_GETSET = 0 
        FLAG_GET_AFTER_SET = 1
        def __init__(self,name,**kwargs):
            pass
        def add_parameter(self,name,**kwargs):
            pass
        def add_function(self,name):
            pass

import visa_prologix as visa
from time import sleep
import sys

class SR_830(Instrument):

    def __init__(self, name, address,ip=""):

        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address,ip=ip)

        self.set_default() # reset everything to:

#REFERENCE / PHASE
#Phase                  0.000 deg
#Reference Source       Internal
#Harmonic               1
#Sine Amplitude         1.000 Vrms
#Internal Frequency     1.000 kHz
#Ext Reference Trigger  Sine
#
#INPUT / FILTERS
#Source                 A
#Grounding              Float
#Coupling               AC
#Line Notches           Out
#
#GAIN / TC
#Sensitivity            1V
#Reserve                Low Noise
#Time Constant          100 ms
#Filter dB/oct.         12 dB
#Synchronous            Off
#
#DISPLAY
#CH1                    X
#CH2                    Y
#Ratio                  None
#Reference              Frequency
#
#OUTPUT / OFFSET
#CH1 Output             X
#CH2 Output             Y
#All Offsets            0.00%
#All Expands            1
#
#AUX OUTPUTS 
#All Output Voltages    0.000 V
#
#SETUP
#Output To              GPIB
#GPIB Address           8
#RS232 Baud Rate        9600
#Parity                 None
#Key Click              On
#Alarms                 On
#Override Remote        On
#
#DATA STORAGE 
#Sample Rate            1 Hz
#Scan Mode              Loop
#Trigger Starts         No
#
#STATUS ENABLE 
#REGISTERS              Cleared

        self.add_function('set_default')
        self.add_function('set_outx')
        self.add_function('get_reference')
        self.add_function('set_reference')
        self.add_function('get_phase_shift')
        self.add_function('set_phase_shift')
        self.add_function('get_frequency')
        self.add_function('set_frequency')
        self.add_function('get_ref_trigger')
        self.add_function('set_ref_trigger')
        self.add_function('get_harmonic')
        self.add_function('set_harmonic')
        self.add_function('get_sine_ampl')
        self.add_function('set_sine_ampl')
        self.add_function('get_input_source')
        self.add_function('set_input_source')
        self.add_function('get_input_shield_gnd')
        self.add_function('set_input_shield_gnd')
        self.add_function('get_input_coupling')
        self.add_function('set_input_coupling')
        self.add_function('get_notch_filters')
        self.add_function('set_notch_filters')
        self.add_function('get_sensitivity')
        self.add_function('set_sensitivity')
        self.add_function('get_reserve_mode')
        self.add_function('set_reserve_mode')
        self.add_function('get_time_constant')
        self.add_function('set_time_constant')
        self.add_function('get_filter_slope')
        self.add_function('set_filter_slope')
        self.add_function('get_sync_filter')
        self.add_function('set_sync_filter')
        self.add_function('get_display_CH1')
        self.add_function('get_display_CH2')
        self.add_function('get_display_CH')
        self.add_function('set_display_CH')
        self.add_function('get_front_panel_output_source_CH1')
        self.add_function('get_front_panel_output_source_CH2')
        self.add_function('get_front_panel_output_source')
        self.add_function('set_front_panel_output_source')
        self.add_function('get_offsets_and_expands_X')
        self.add_function('get_offsets_and_expands_Y')
        self.add_function('get_offsets_and_expands_R')
        self.add_function('get_offsets_and_expands')
        self.add_function('set_offsets_and_expands')
        self.add_function('set_auto_offset_X')
        self.add_function('set_auto_offset_Y')
        self.add_function('set_auto_offset_R')
        self.add_function('set_auto_offset')
        self.add_function('get_aux_input_AUX1')
        self.add_function('get_aux_input_AUX2')
        self.add_function('get_aux_input_AUX3')
        self.add_function('get_aux_input_AUX4')
        self.add_function('get_aux_input')
        self.add_function('get_aux_output')
        self.add_function('get_aux_output_AUX1')
        self.add_function('get_aux_output_AUX2')
        self.add_function('get_aux_output_AUX3')
        self.add_function('get_aux_output_AUX4')
        self.add_function('set_aux_output')
        self.add_function('set_override_remote')
        self.add_function('save_setup')
        self.add_function('recall_setup')
        self.add_function('get_output')
        self.add_function('get_output_X')
        self.add_function('get_output_Y')
        self.add_function('get_output_Z')
        self.add_function('get_output_theta')
        self.add_function('get_snap')

    # communication with device

    def set_default(self):
        com_str = "*RST"
        self._visainstrument.write(com_str)
        sleep(2)
        return 0

    def set_outx(self,i = 1): # RS232 (i=0) or GPIB (i=1)
        i = int(bool(i)) # bam! (forces everything to be zero or one)
        com_str = "OUTX %i" % i
        self._visainstrument.write(com_str)
        sleep(0.05)
        return i

    def get_reference(self):
        self.set_outx(1)
        com_str = "FMOD?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_reference(self, ref = 0): ## reference? 1 = internal, 0 = external
        if (ref != 1) and (ref != 0):
            print >> sys.stderr, "Input error in set_reference()!"
            return -1
        else:
            com_str = ("FMOD %i") % ref
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_reference()

    def get_phase_shift(self):
        self.set_outx(1)
        com_str = "PHAS?"
        self._visainstrument.write(com_str)
        return float(self._visainstrument.read())

    def set_phase_shift(self,phas = 0.0):
        phas = float(phas)
        if (phas < -360.00) or (phas > 729.99):
            print >> sys.stderr, "Input error in set_phase_shift()!"
            return -1
        else:
            com_str = "PHAS %.2f" % phas
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_phase_shift()

    def get_frequency(self):
        self.set_outx(1)
        com_str = "FREQ?"
        self._visainstrument.write(com_str)
        return float(self._visainstrument.read())

    def set_frequency(self, freq = 102000):
        freq = float(freq)
        if (freq < 0.001) or (freq > 102000):
            print >> sys.stderr, "Input error in set_frequency()!"
            return -1
        elif self.get_reference() != 1:
            print >> sys.stderr, "Error in set_frequency(): SR830 needs to be set to internal reference!"
            return -1
        else:
            com_str = "FREQ %.4f" % freq
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_frequency()

    def get_ref_trigger(self): # 0 = sine zero crossing, 1 = TTL rising, 2 = TTL falling
        self.set_outx(1)
        com_str = "RSLP?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_ref_trigger(self, ref = 1):
        if (ref != 1) and (ref != 0) and (ref != 2):
            print >> sys.stderr, "Input error in set_ref_trigger()!"
            return -1
        elif self.get_reference() != 0:
            print >> sys.stderr, "Error in ref_trigger(): SR830 needs to be set to external reference!"
            return -1
        else:
            com_str = ("RSLP %i") % ref
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_ref_trigger()

    def get_harmonic(self):
        self.set_outx(1)
        com_str = "HARM?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_harmonic(self, harm = 1):
        harm = int(harm)
        if (harm < 1) or (harm > 19999):
            print >> sys.stderr, "Input error in set_harmonic()!"
            return -1
        else: 
            com_str = "HARM %i" % harm
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_harmonic()

    def get_sine_ampl(self):
        self.set_outx(1)
        com_str = "SLVL?"
        self._visainstrument.write(com_str)
        return float(self._visainstrument.read())

    def set_sine_ampl(self, ampl = 0.004):
        ampl = float(ampl)
        if (ampl < 0.004) or (ampl > 5.0):
            print >> sys.stderr, "Input error in set_sine_ampl()!"
            return -1
        else:
            com_str = "SLVL %f" % ampl
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_sine_ampl()

    def get_input_source(self):     # A (i=0) single ended, A-B (i=1) differential, 
        self.set_outx(1)            # I (1 MOhm) (i=2) or I (100 MOhm) (i=3)
        com_str = "ISRC?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_input_source(self, isrc = 0):
        isrc = int(isrc)
        if (isrc < 0) or (isrc > 3):
            print >> sys.stderr, "Input error in set_input_source()!"
            return -1
        else:
            com_str = "ISRC %i" % isrc
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_input_source()

    def get_input_shield_gnd(self): # floating (0) or grounded (1)
        self.set_outx(1)
        com_str = "IGND?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_input_shield_gnd(self, gnd = 0):
        gnd = int(gnd)
        if (gnd < 0) or (gnd > 1):
            print >> sys.stderr, "Input error in set_input_shield_gnd()!"
            return -1
        else:
            com_str = "IGND %i" % gnd
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_input_shield_gnd()

    def get_input_coupling(self): # AC (0) or DC (1)
        self.set_outx(1)
        com_str = "ICPL?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_input_coupling(self, icpl = 0):
        icpl = int(icpl)
        if (icpl < 0) or (icpl > 1):
            print >> sys.stderr, "Input error in set_input_coupling()!"
            return -1
        else:
            com_str = "ICPL %i" % icpl
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_input_coupling()

    def get_notch_filters(self):
        self.set_outx(1)
        com_str = "ILIN?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_notch_filters(self, ilin = 3): # no filters (i=0), Line notch in (i=1), 
        ilin = int(ilin)                   # 2xLine notch in (i=2) or Both notch filters in (i=3)    
        if (ilin < 0) or (ilin > 3):
            print >> sys.stderr, "Input error in set_notch_filters()!"
            return -1
        else:
            com_str = "ILIN %i" % ilin
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_notch_filters()

    def get_sensitivity(self):
        self.set_outx(1)
        com_str = "SENS?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    # i   sensitivity
    # 0   2 nV/fA
    # 1   5 nV/fA
    # 2   10 nV/fA
    # 3   20 nV/fA
    # 4   50 nV/fA
    # 5   100 nV/fA
    # 6   200 nV/fA
    # 7   500 nV/fA
    # 8   1 muV/pA
    # 9   2 muV/pA
    # 10  5 muV/pA
    # 11  10 muV/pA
    # 12  20 muV/pA
    # 13  50 muV/pA
    # 14  100 muV/pA
    # 15  200 muV/pA
    # 16  500 muV/pA
    # 17  1 mV/nA
    # 18  2 mV/nA
    # 19  5 mV/nA
    # 20  10 mV/nA
    # 21  20 mV/nA
    # 22  50 mV/nA
    # 23  100 mV/nA
    # 24  200 mV/nA
    # 25  500 mV/nA
    # 26  1 V/muA

    def set_sensitivity(self, sens = 10):
        sens = int(sens)
        if (sens < 0) or (sens > 26):
            print >> sys.stderr, "Input error in set_sensitivity()!"
            return -1
        else:
            com_str = "SENS %i" % sens
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_sensitivity()

    def get_reserve_mode(self):  # high reserve (0), normal (1), low noise (2)
        self.set_outx(1)
        com_str = "RMOD?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_reserve_mode(self, rmod = 1):
        rmod = int(rmod)
        if (rmod < 0) or (rmod > 2):
            print >> sys.stderr, "Input error in set_reserve_mode()!"
            return -1
        else:
            com_str = "RMOD %i" % rmod
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_reserve_mode()

    def get_time_constant(self):
        self.set_outx(1)
        com_str = "OFLT?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    # i time constant
    # 0     10 mus
    # 1     30 mus
    # 2     100 mus
    # 3     300 mus
    # 4     1 ms
    # 5     3 ms
    # 6     10 ms
    # 7     30 ms
    # 8     100 ms
    # 9     300 ms
    # 10    1s
    # 11    3s
    # 12    10 s
    # 13    30 s
    # 14    100 s
    # 15    300 s
    # 16    1 ks
    # 17    3 ks
    # 18    10 ks
    # 19    30 ks

    def set_time_constant(self, timec = 7):
        timec = int(timec)
        if (timec < 0) or (timec > 19):
            print >> sys.stderr, "Input error in set_time_constant()!"
            return -1
        else:
            com_str = "OFLT %i" % timec
            self._visainstrument.write(com_str)
            sleep(1)
            actual_time_constant = self.get_time_constant()
            if (actual_time_constant != timec):
                print >> sys.stderr, "Warning: time constant could not be set (incompatible with filters?)!"
            return actual_time_constant

    def get_filter_slope(self):
        self.set_outx(1)
        com_str = "OFSL?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    # 0      6 dB/oct
    # 1     12 dB/oct
    # 2     18 dB/oct
    # 3     24 dB/oct

    def set_filter_slope(self, slp = 1):
        slp = int(slp)
        if (slp < 0) or (slp > 3):
            print >> sys.stderr, "Input error in set_filter_slope()!"
            return -1
        else:
            com_str = "OFSL %i" % slp
            self._visainstrument.write(com_str)
            sleep(0.2)
            return self.get_filter_slope()

    def get_sync_filter(self):  # off (0) or synchronous filter below 200 Hz (1)
        self.set_outx(1)
        com_str = "SYNC?"
        self._visainstrument.write(com_str)
        return int(self._visainstrument.read())

    def set_sync_filter(self, sync = 0):
        sync = int(sync)
        if (sync < 0) or (sync > 1):
            print >> sys.stderr, "Input error in set_sync_filter()!"
            return -1
        else:
            if (sync == 1) and ((self.get_harmonic()*self.get_frequency()) > 200):
                print >> sys.stderr, "No synchronous filtering: detection frequency above 200 Hz!"
                return -1
            else:
                com_str = "SYNC %i" % sync
                self._visainstrument.write(com_str)
                sleep(0.05)
                return self.get_sync_filter()

    def get_display_CH1(self):
        self.set_outx(1)
        com_str = "DDEF? 1"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(res_str.split(',')[0]),int(res_str.split(',')[1])

    def get_display_CH2(self):
        self.set_outx(1)
        com_str = "DDEF? 2"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(res_str.split(',')[0]),int(res_str.split(',')[1])

    def get_display_CH(self, number = 1):
        number = int(number)
        if number == 1:
            return self.get_display_CH1()
        elif number == 2:
            return self.get_display_CH2()
        else:
            print >> sys.stderr, "get_display_CH(): invalid display number!"
            return -1

    # CH1 (1)         CH2 (2)
    #
    # display
    # 0   X           0   Y
    # 1   R           1   theta
    # 2   X noise     2   Y noise
    # 3   Aux In 1    3   Aux In 3
    # 4   Aux In 2    4   Aux In 4
    #
    # ratio
    # 0   none        0   none
    # 1   Aux In 1    1   Aux In 3
    # 2   Aux In 2    2   Aux In 4

    def set_display_CH(self, chan = 1, display = 0, ratio = 0):
        chan = int(chan)
        display = int(display)
        ratio = int(ratio)
        if (chan < 1) or (chan > 2):
            print >> sys.stderr, "Input error in set_display_CH()!"
            return -1
        elif (display < 0) or (display > 4):
            print >> sys.stderr, "Input error in set_display_CH()!"
            return -1
        elif (ratio < 0) or (ratio > 2):
            print >> sys.stderr, "Input error in set_display_CH()!"
            return -1
        else:
            com_str = "DDEF %i,%i,%i" % (chan, display, ratio)
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_display_CH(chan)

    def get_front_panel_output_source_CH1(self):
        self.set_outx(1)
        com_str = "FPOP? 1"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(res_str)

    def get_front_panel_output_source_CH2(self):
        self.set_outx(1)
        com_str = "FPOP? 2"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(res_str)

    def get_front_panel_output_source(self, number = 1):
        number = int(number)
        if number == 1:
            return self.get_front_panel_output_source_CH1()
        elif number == 2:
            return self.get_front_panel_output_source_CH2()
        else:
            print >> sys.stderr, "get_front_panel_output_source(): invalid display number!"
            return -1

    #     CH1 (1)       CH2 (2)
    #    output quantity
    # 0   CH1 Display   CH2 Display
    # 1   X             Y

    def set_front_panel_output_source(self, chan = 1, fpop = 0):
        chan = int(chan)
        fpop = int(fpop)
        if (chan < 1) or (chan > 2):
            print >> sys.stderr, "Input error in set_display_CH()!"
            return -1
        elif (fpop < 0) or (fpop > 1):
            print >> sys.stderr, "Input error in set_display_CH()!"
            return -1
        else:
            com_str = "FPOP %i,%i" % (chan, fpop)
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_front_panel_output_source(chan)

    def get_offsets_and_expands_X(self):
        self.set_outx(1)
        com_str = "OEXP? 1"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str.split(',')[0]),int(res_str.split(',')[1])

    def get_offsets_and_expands_Y(self):
        self.set_outx(1)
        com_str = "OEXP? 2"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str.split(',')[0]),int(res_str.split(',')[1])

    def get_offsets_and_expands_R(self):
        self.set_outx(1)
        com_str = "OEXP? 3"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str.split(',')[0]),int(res_str.split(',')[1])

    def get_offsets_and_expands(self, number = 1):
        number = int(number)
        if number == 1:
            return self.get_offsets_and_expands_X()
        elif number == 2:
            return self.get_offsets_and_expands_Y()
        elif number == 3:
            return self.get_offsets_and_expands_R()
        else:
            print >> sys.stderr, "get_front_panel_output_source(): invalid display number!"
            return -1

    # first parameter is X (1), Y (2), R (3)
    # second parameter is offset in percent (-105.00 <= x <= 105.00)
    # third parameter is expand - 0:no expand, 1:expand by 10, 2:expand by 100

    def set_offsets_and_expands(self, chan = 1, offset = 0.0, expand = 0):
        chan = int(chan)
        offset = float(offset)
        expand = int(expand)
        if (chan < 1) or (chan > 3):
            print >> sys.stderr, "Input error in set_offsets_and_expands()!"
            return -1
        if abs(offset) > 105.0:
            print >> sys.stderr, "Input error in set_offsets_and_expands()!"
            return -1
        if (expand < 0) or (expand > 2):
            print >> sys.stderr, "Input error in set_offsets_and_expands()!"
            return -1
        else:
            com_str = "OEXP %i,%f,%i" % (chan, offset, expand)
            self._visainstrument.write(com_str)
            sleep(0.5)
            return self.get_offsets_and_expands(chan)

    def set_auto_offset_X(self):  # turn off auto offset with:
        com_str = "AOFF 1"    # set_offsets_and_expands(i, 0.0, k)
        self._visainstrument.write(com_str)
        return 0

    def set_auto_offset_Y(self):
        com_str = "AOFF 2"
        self._visainstrument.write(com_str)
        return 0

    def set_auto_offset_R(self):
        com_str = "AOFF 3"
        self._visainstrument.write(com_str)
        return 0

    def set_auto_offset(self, chan = 1):
        chan = int(chan)
        if chan == 1:
            return self.set_auto_offset_X()
        elif chan == 2:
            return self.set_auto_offset_Y()   
        elif chan == 3:
            return self.set_auto_offset_R()
        else:     
            print >> sys.stderr, "set_auto_offset(): invalid variable (X=1, Y=2, R=3)!"
            return -1

    def get_aux_input_AUX1(self): # in units of volt
        self.set_outx(1)
        com_str = "OAUX? 1"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_aux_input_AUX2(self):
        self.set_outx(1)
        com_str = "OAUX? 2"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_aux_input_AUX3(self):
        self.set_outx(1)
        com_str = "OAUX? 3"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_aux_input_AUX4(self):
        self.set_outx(1)
        com_str = "OAUX? 4"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_aux_input(self, aux = 1):
        aux = int(aux)
        if aux == 1:
            return self.get_aux_input_AUX1()
        elif aux == 2:
            return self.get_aux_input_AUX2()
        elif aux == 3:
            return self.get_aux_input_AUX3()
        elif aux == 4:
            return self.get_aux_input_AUX4()
        else:
            print >> sys.stderr, "get_aux_input(): invalid channel (choose 1-4)!"
            return -1

    def get_aux_output(self, chan = 1):
        chan = int(chan)
        if (chan < 1) or (chan > 4):
            print >> sys.stderr, "get_aux_output(): invalid channel (choose 1-4)!"
            return -1
        else:
            self.set_outx(1)
            com_str = "AUXV? %i" % chan
            self._visainstrument.write(com_str)
            res_str = self._visainstrument.read()
            return float(res_str)

    def get_aux_output_AUX1(self):
        return self.get_aux_output(1)

    def get_aux_output_AUX2(self):
        return self.get_aux_output(2)

    def get_aux_output_AUX3(self):
        return self.get_aux_output(3)

    def get_aux_output_AUX4(self):
        return self.get_aux_output(4)

    def set_aux_output(self, chan = 1, auxv = 0.0):
        chan = int(chan)
        auxv = float(auxv)
        if (chan < 1) or (chan > 4):
            print >> sys.stderr, "set_aux_output(): invalid channel (choose 1-4)!"
            return -1
        elif abs(auxv) > 10.5:
            print >> sys.stderr, "set_aux_output(): voltage too large!"
            return -1
        else:
            com_str = "AUXV %i,%f" % (chan,auxv)
            self._visainstrument.write(com_str)
            sleep(0.05)
            return self.get_aux_output(chan)

    def set_override_remote(self, mode = 0): # 1: override GPIB remote mode; 0: normal operation (GPIB 
        mode = int(bool(mode))               # blocks front panel access)
        com_str = "OVRM %i" % mode
        self._visainstrument.write(com_str)
        sleep(0.05)
        return mode

################  below: storage not propperly tested  ################

    def save_setup(self, n_buffer = 1):  # Save setup in a buffer. Buffers will survive power off.
        n_buffer = int(n_buffer)
        if (n_buffer < 1) or (n_buffer > 9):
            print >> sys.stderr, "save_setup(): invalid buffer (choose 1-9)!"
            return -1
        else:
            com_str = "SSET %i" % n_buffer
            self._visainstrument.write(com_str)
            sleep(0.05)
            return n_buffer

    def recall_setup(self, n_buffer = 1):  # Recall setup from buffer.
        n_buffer = int(n_buffer)
        if (n_buffer < 1) or (n_buffer > 9):
            print >> sys.stderr, "recall_setup(): invalid buffer (choose 1-9)!"
            return -1
        else:
            com_str = "RSET %i" % n_buffer
            self._visainstrument.write(com_str)
            sleep(0.5)
            return n_buffer

################  above: storage not propperly tested  ################

    def to_seconds(self, time_const):  # decodes the time constant setting to seconds
        time_const = int(time_const)
        if (time_const%2 == 1):
            factor = 3.0
        else:
            factor = 1.0
        return factor*0.001*0.01*pow(10,time_const/2) 

# since I don't check the status bit, the auto functions are too unstable
#    def set_auto_gain(self):
#        time_const = self.get_time_constant()
#        if time_const > 7:
#            print >> sys.stderr, "set_auto_gain(): not possible, time constant too high."
#            return -1
#        com_str = "AGAN"
#        self._visainstrument.write(com_str)
#        sleep(max(20*self.to_seconds(time_const),1.0))  # auto gain takes surprisingly long
#        self.get_time_constant() # just to force a crash here if the instrument still 
#        return 0                 # does not react after 5 seconds of sleep
#
#    def set_auto_reserve(self):
#        self.get_time_constant()
#        com_str = "ARSV"
#        self._visainstrument.write(com_str)
#        sleep(10)
#        self.get_time_constant() # just to force a crash here if the instrument still 
#        return 0                 # does not react after 5 seconds of sleep
#
# will not include the auto phase here, since it "takes many time constants" to apply

    # finally, to measuring

    def get_output(self, meas = 1): # 1:X, 2:Y, 3:R, 4:theta
        meas = int(meas)
        if (meas < 1) or (meas > 4):
            print >> sys.stderr, "Input error in get_output(): choose 1-4"
            return -1
        else:
            self.set_outx(1)
            com_str = "OUTP? %i" % meas
            self._visainstrument.write(com_str)
            res_str = self._visainstrument.read()
            return float(res_str)

    def get_output_X(self):
        return self.get_output(1)

    def get_output_Y(self):
        return self.get_output(2)

    def get_output_Z(self):
        return self.get_output(3)

    def get_output_theta(self):
        return self.get_output(4)

    def get_snap(self, nargs=2, query_str="1,2"): # simultanious reading of several parameters
        nargs = int(nargs)
        query_str = query_str.split(',')
        if len(query_str) < nargs or nargs < 2 or nargs > 6:
            print >> sys.stderr, "Input error in get_snap()!"    
            return -1
        com_str = "SNAP? "
        for i in range(nargs):
            temp= int(query_str[i])
            if (temp < 1) or (temp > 11):
                print >> sys.stderr, "Input error in get_snap()!"    
                return -1
            else:
                com_str = "%s%i," % (com_str,temp)
        self.set_outx(1)
        com_str = com_str[0:(len(com_str)-1)]
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()        
        res_str = res_str.split(',')
        for i in range(nargs):
            res_str[i] = float(res_str[i])
        return res_str

    # The SNAP? command requires at least two parameters and at most six
    # parameters. The parameters i, j, k, l, m, n select the parameters below.
    # 1     X
    # 2     Y
    # 3     R
    # 4     theta
    # 5     Aux In 1
    # 6     Aux In 2
    # 7     Aux In 3
    # 8     Aux In 4
    # 9     Reference Frequency
    # 10    CH1 display
    # 11    CH2 display


## testing:

if __name__ == "__main__":
    sr830=SR_830("SR830","GPIB::8",ip="192.168.0.100")

    sr830._visainstrument.write("*IDN?")
    print "IDN: ", sr830._visainstrument.read()

#    print sr830.set_time_constant(7)

    print "Freq:", sr830.get_frequency()

    print "Sync:", sr830.get_sync_filter()

    print sr830.set_frequency(150)

    print sr830.set_sync_filter(1)

    print "Freq:", sr830.get_frequency()

    print "Sync:", sr830.get_sync_filter()

    print sr830.get_input_source()

    print sr830.get_snap(4, "1,2,6,7")[2]



# filename: Virtual_Current_Source.py by JB@KIT 05/2018
'''
Virtual instrument to be used as a wrapper for the DC DAC LTC2666. 
The wrapper is used with a current source that outputs a current proportional to an 
input voltage, which is in turn controlled by the DC DAC LTC2666. 
The LTC2666 driver is called DC_DAC_LTC2666.py. 

Use:
  - set current range of current source (proportionality constant)
  - current unit is mA, independent of the specified range

Import and usage:

vcurr = qkit.instruments.create('vcurr', 'Virtual_Current_Source', host='129.00.00.00', port=9999)
vcurr.set_ch0_range(1)
vcoil.set_current(1.0)

Current ranges are given as a float number such that the resulting voltage is 
voltage = current * current_range e.g. 

current_range=1   ->    1V ^= 1mA
current_range=10  ->    1V ^= 10mA
current_range=0.01  ->  1V ^= 10uA

The DAC is 16bit, so the voltage resolution is 10V / 65536 = 0.15mV
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit.instruments.DC_DAC_LTC2666 import DC_DAC_LTC2666
import logging
import types
import time
from qkit.gui.notebook.Progress_Bar import Progress_Bar
import numpy as np

class Virtual_Current_Source(DC_DAC_LTC2666):
    '''
    Instrument wrapper for DC DAC LTC2666, used as a set of dc current sources. 
    Inherits from DC_DAC_LTC2666.
    '''
    def __init__(self, name, host='ip-address', port=9931, nchannels = 16):
        '''
        This init method directly creates current source instruments, instead of voltage sources.
        '''
        Instrument.__init__(self, name, tags=['virtual'])

        self.channels = nchannels
        self.add_parameter('current', type=float, flags=Instrument.FLAG_GETSET, 
                units='mA', channels=(0, self.channels-1), channel_prefix='ch%d_')
        self.add_parameter('range', type=float, flags=Instrument.FLAG_GETSET, 
                channels=(0, self.channels-1), channel_prefix='ch%d_')
        self.currents = [0]*self.channels
        self.c_ranges = [1]*self.channels
            
        self.HOST, self.PORT = host, port
        self.init_connection()   #DC_DAC_LTC2666.py

    def do_set_range(self, crange, channel = 0):
        """
        Set the current range of channel. This does not change the current output directly 
        but takes effect only at setting a new current.
        """
        self.c_ranges[channel] = crange
        logging.info(__name__ + ' : Setting current range of channel {:d} to {:f}.'.format(channel, crange))

    def do_get_range(self, channel = 0):
        """
        Get the current range of channel.
        """
        return self.c_ranges[channel]

    def do_set_current(self, current, channel = 0, verbose=True):
        """
        Set current of channel in mA. 
        """
        try:
            voltage = current * self.c_ranges[channel]
            if voltage < -5. or voltage >= 5.:
                raise ArithmeticError
            else:
                self.do_set_voltage(voltage, channel)   #DC_DAC_LTC2666.py
                self.currents[channel] = current
                if verbose:
                    logging.info(__name__ + ' : Setting current of channel {:d} to {:f}mA.'.format(channel, current))
                
        except ArithmeticError as detail:
            logging.error('Invalid current setting, value exceeds voltage threshold! No changees made.')
            print detail

    def do_get_current(self, channel):
        '''
        Get the current current setting for channel.
        '''
        return self.currents[channel]
            
    def ramp_to(self, target = 0, channel = 0, steps = 20, dt = 0.1):
        """
        Ramp current to target
        Inputs:
            - target: target current value (mA)
            - channel: channel index (1..)
            - steps: number of steps
            - dt: wait time between two steps
        """
        p = Progress_Bar(steps,'Ramping current')
        for c in np.linspace(self.do_get_current(channel),target,steps):
            self.do_set_current(c, channel, verbose=False)
            p.iterate('{:.3g}mA'.format(c))
            time.sleep(dt)
        self.do_set_current(target, channel ,verbose=True)




"""
sequency_library.py
A. Stehli (05/2018)

This library provides the measurement sequences for standard qubit experiments.
Currently this includes:
    rabi experiment
    T1 measurments
    Ramsey experiments
    Spin/Hahn echo
"""

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit.measure.timedomain.pulse_sequence as ps

"""
TODO: How to implement the mixer calibration? Should it be stored in the sample object or in an external file where it can be acessed?
"""


def rabi(sample, pulse_shape=ps.ShapeLib.rect, amplitude=1, iq_frequency=None):
    """
        Generate sequence object for a rabi experiment (varying drive pulse lengths).

        Args:
            sample:      sample object
            pulse_shape: shape of the pulses (i.e. square, gauss, ...)
            amplitude:   relative ampitude of the sequence
            iq_frequency: IQ-frequency for heterodyne mixing. If iq_frequency is None, the sample.iq_frequency is used.
                            If iq_frequency is 0 homodyne mixing is employed.

        Returns:
            Sequence object for rabi experiment
    """
    if iq_frequency is None:
        if hasattr(sample, "iq_frequency"):
            iq_frequency = sample.iq_frequency
            print("IQ-frequency set to {:.0f} MHz.".format(iq_frequency / 1e6))
        else:
            iq_frequency = 0
            print(
                "Sample has no attribute iq_frequency.\n IQ-frequency set to 0 for homodyne mixing.")
    rabi_tone = ps.Pulse(lambda t: t, shape=pulse_shape, name="rabi-tone",
                         amplitude=amplitude, iq_frequency=iq_frequency)
    sequence = ps.PulseSequence(sample)
    sequence.add(rabi_tone)
    sequence.add_readout()
    return sequence


def t1(sample, pulse_shape=ps.ShapeLib.rect, amplitude=1, iq_frequency=None):
    """
        Generate a sequence with one pi pulse followed by a time delay.

        Args:
            sample:       sample object
            pulse_shape:  shape of the pulses (i.e. square, gauss, ...)
            amplitude:    relative ampitude of the sequence
            iq_frequency: IQ-frequency for heterodyne mixing. If iq_frequency is None, the sample.iq_frequency is used.
                            If iq_frequency is 0 homodyne mixing is employed.

        Returns:
            Sequence object for T1 measurement
    """
    if iq_frequency is None:
        if hasattr(sample, "iq_frequency"):
            iq_frequency = sample.iq_frequency
            print("IQ-frequency set to {:.0f} MHz.".format(iq_frequency / 1e6))
        else:
            iq_frequency = 0
            print(
                "Sample has no attribute iq_frequency.\n IQ-frequency set to 0 for homodyne mixing.")
    pi_pulse = ps.Pulse(sample.tpi, shape=pulse_shape, name="pi",
                        amplitude=amplitude, iq_frequency=iq_frequency)
    sequence = ps.PulseSequence(sample)
    sequence.add(pi_pulse)
    sequence.add_wait(lambda t: t)
    sequence.add_readout()
    return sequence


def ramsey(sample, pulse_shape=ps.ShapeLib.rect, amplitude=1, iq_frequency=None):
    """
    Generate a sequence with two pi/2 pulses seperated by a time delay.

    Args:
        sample:       sample object
        pulse_shape:  shape of the pulses (i.e. square, gauss, ...)
        amplitude:    relative ampitude of the sequence
        iq_frequency: IQ-frequency for heterodyne mixing. If iq_frequency is None, the sample.iq_frequency is used.
                      If iq_frequency is 0 homodyne mixing is employed.

    Returns:
        Sequence object for ramsey measurement
    """
    return spinecho(sample, 0, pulse_shape, amplitude, iq_frequency)


def spinecho(sample, n_pi=1, pulse_shape=ps.ShapeLib.rect, amplitude=1, iq_frequency=None):
    """
    Generate sequence with two pi/2 pulses at the ends and a number n_pi pi pulses inbetween
    pi2 - time/(2*n) - pi - time/n - pi - ... - pi - time/(2*n) - pi2


    Args:
        sample:       sample object
        n_pi:         number of pi-pulses
        pulse_shape:  shape of the pulses (i.e. square, gauss, ...)
        amplitude:    relative ampitude of the sequence
        iq_frequency: IQ-frequency for heterodyne mixing. If iq_frequency is None, the sample.iq_frequency is used.
                      If iq_frequency is 0 homodyne mixing is employed.

    Returns:
        Sequence object for spinecho measurement
    """
    if iq_frequency is None:
        if hasattr(sample, "iq_frequency"):
            iq_frequency = sample.iq_frequency
            print("IQ-frequency set to {:.0f} MHz.".format(iq_frequency / 1e6))
        else:
            iq_frequency = 0
            print(
                "Sample has no attribute iq_frequency.\n IQ-frequency set to 0 for homodyne mixing.")
    pi_pulse = ps.Pulse(sample.tpi, shape=pulse_shape, name="pi",
                        amplitude=amplitude, iq_frequency=iq_frequency)
    pi2_pulse = ps.Pulse(sample.tpi2, shape=pulse_shape, name="pi/2",
                         amplitude=amplitude, iq_frequency=iq_frequency)

    sequence = ps.PulseSequence(sample)
    sequence.add(pi2_pulse)
    if n_pi == 0:
        sequence.add_wait(lambda t: t)
    else:
        sequence.add_wait(lambda t: t/(2*n_pi))
        for i in range(n_pi):
            sequence.add(pi_pulse)
            if i + 1 != n_pi:
                sequence.add_wait(lambda t: t/n_pi)
        sequence.add_wait(lambda t: t/(2*n_pi))
    sequence.add(pi2_pulse)
    sequence.add_readout()
    return sequence


def spinlocking(sample, pi2sign=1, add_pi=True, wait_time=5e-9, pulse_shape=ps.ShapeLib.rect, amplitude=1, iq_frequency=None):
    """
    Generate sequence with two pi/2 pulses (around y-axis) at the ends and a drive in-between.
    The drive is phase shifted by 90 degrees in respect to the pi/2 pulses (rotates around x-axis).
    Furthermore a pi-pulse (around x-axis) may be added (True by default) before as well as after the drive.
    This reduces the susceptibility to low frequency oscillations.

    Args:
        sample:       sample object
        pi2sign:      sign of the first pi/2 pulse
        add_pi:       if True an additional pi-pulse is added before and after the drive
        wait_time:    wait time between pi/2 or pi-pulse and the drive
        pulse_shape:  shape of the pi and pi/2 pulses (i.e. square, gauss, ...)
        amplitude:    relative ampitude of the sequence
        iq_frequency: IQ-frequency for heterodyne mixing. If iq_frequency is None, the sample.iq_frequency is used.
                        If iq_frequency is 0 homodyne mixing is employed.

    Returns:
        Sequence object for spinecho measurement
    """
    if iq_frequency is None:
        if hasattr(sample, "iq_frequency"):
            iq_frequency = sample.iq_frequency
            print("IQ-frequency set to {:.0f} MHz.".format(iq_frequency / 1e6))
        else:
            iq_frequency = 0
            print(
                "Sample has no attribute iq_frequency.\n IQ-frequency set to 0 for homodyne mixing.")
    # check if pi/2-pulse needs to be flipped
    if pi2sign == -1:
        phase = 270
    else:
        phase = 90

    # define pulses
    pi_pulse = ps.Pulse(sample.tpi, shape=pulse_shape, name="pi",
                        amplitude=amplitude, iq_frequency=iq_frequency)
    pi2_pulse1 = ps.Pulse(sample.tpi2, shape=pulse_shape, name="pi/2",
                          amplitude=amplitude, iq_frequency=iq_frequency, phase=phase)
    pi2_pulse2 = ps.Pulse(sample.tpi2, shape=pulse_shape, name="pi/2(Y)",
                          amplitude=amplitude, iq_frequency=iq_frequency, phase=90)
    locking_tone = ps.Pulse(lambda t: t, shape=ps.ShapeLib.rect,
                            name="locking-drive", amplitude=amplitude, iq_frequency=iq_frequency)

    # add pulses to sequence
    sequence = ps.PulseSequence(sample)
    sequence.add(pi2_pulse1)
    if add_pi:
        sequence.add(pi_pulse)
    sequence.add_wait(wait_time)
    sequence.add(locking_tone)
    sequence.add_wait(wait_time)
    if add_pi:
        sequence.add(pi_pulse)
    sequence.add(pi2_pulse2)
    sequence.add_readout()
    return sequence



import numpy as np


def seconds_to_str(secs):
    '''Convert a number of seconds to hh:mm:ss string.'''
    hours = np.floor(secs / 3600)
    secs -= hours * 3600
    mins = np.floor(secs / 60)
    secs = np.floor(secs - mins * 60)
    return '%02d:%02d:%02d' % (hours, mins, secs)

def get_ipython():
    import IPython
    if ipython_is_newer((0, 11)):
        return IPython.get_ipython()
    else:
        return IPython.ipapi.get()

def get_traceback():
    if ipython_is_newer((0, 11)):
        from IPython.core.ultratb import AutoFormattedTB
    else:
        from IPython.ultraTB import AutoFormattedTB
    return AutoFormattedTB

def ipython_is_newer(vin):
    """
    vin is tuple of version (a,b,c) for version "a.b.c"
    result gives True for larger or equal version
    """
    import IPython
    vs = IPython.__version__.split('.')
    for i in range(len(vs)):
        if i > (len(vin)-1):
            return True
        if int(vs[i]) > vin[i]:
            return True
        elif int(vs[i]) < vin[i]:
            return False
    return True

def is_ipython():
    return get_ipython() != None

def register_exit(func):
    if is_ipython():
        ip = get_ipython()
        if ipython_is_newer((8, 0, 0)):
            import atexit
            atexit.register(func)
        elif ipython_is_newer((0, 11)):
            ip.hooks['shutdown_hook'].add(func, 1)
        else:
            ip.IP.hooks.shutdown_hook.add(func, 1)
    else:
        import atexit
        atexit.register(func)

def str3(string,encoding='UTF-8'):
    try:
        return string.decode(encoding)
    except AttributeError:
        return str(string)

def concat(*args):
    """
    Concatenates any number of strings or bytestrings together to a string.
    This function can be helpful for code migration from py2
    """
    s = ""
    for a in args:
        s+=str3(a)
    return s



from __future__ import print_function
# start zmq based info_service
# on port and host defined in config/environment
# HR@KIT/2017
import qkit

if qkit.cfg.get('load_info_service',True):
    qkit.cfg['load_info_service']=True
    try: 
        import zmq
    except ImportError:
        print("Please install the zmq package")
        # we handle this exception again in the module
    
    from qkit.core.lib.com.info_service import info_service
    qkit.info = info_service()
else:
    # dummy info service
    def info_service(msg): pass
    qkit.info = info_service



# -*- coding: utf-8 -*-
"""
Library to ease the use of the file format hdf5 for datastorage
It can be used with or without the qtlab environment.

@author: hannes.rotzinger@kit.edu 2018
@author: marco.pfirrmann@kit.edu 2018
@version: 0.1
"""
import logging
import os
import traceback

import qkit
from qkit.storage.hdf_file import H5_file
from qkit.storage.hdf_dataset import hdf_dataset
from qkit.storage.hdf_constants import ds_types
from qkit.storage.hdf_view import dataset_view
from qkit.storage.hdf_DateTimeGenerator import DateTimeGenerator



class Data(object):
    """Basic hdf5 class adopted to our needs.
    
    This class is the public API for the qkit users with the hdf5 file. It
    is based in its core on h5py as well as our own adaptions in hdf_view, 
    hdf_file, and hdf_dataset. Here mostly have wrapper that operate on the
    mentioned classes.
    """
    # a types
    def __init__(self, name = None, mode = 'r+', copy_file = False):
        """Creates an empty data set including the file, for which the currently
        set file name generator is used or opens the h5 file at location 'name'.

        Args:
            name (string):  filename or absolute filepath
            mode (string):  access mode to the hdf5 file, default: 'r+' (read+write).
                Other modes are 'a' (read, write, and create)
        """
        self._name = name
        if os.path.isfile(self._name):
            self._filepath = os.path.abspath(self._name)
            self._folder, self._filename = os.path.split(self._filepath)
        elif not os.path.isabs(self._name):
            self._generate_file_name()
            try:
                qkit.fid.add_h5_file(self._filepath)
            except Exception as e:
                logging.debug("Could not add newly generated h5 File '{}' to qkit.fid database: {}".format(name,e))
        else:
            self._filepath = os.path.abspath(self._name)
            self._folder,self._filename = os.path.split(self._filepath)
        "setup the  file"
        try:
            self.hf = H5_file(self._filepath, mode)
        except IOError:
            raise IOError('File does not exist. Use argument \"mode=\'a\'\" to create a new h5 file.')
        if self.hf.newfile:
            if self.__dict__.get('_uuid', False):
                tags = ["_unix_timestamp", "_localtime", "_timestamp", "_timemark", "_datemark", "_uuid", "_filename", "_folder", "_relpath", "_filepath"]
                self.hf.hf.attrs.update({a: self.__dict__.get(a) for a in tags})
            if "user" in qkit.cfg:
                self.hf.hf.attrs['_user'] = qkit.cfg.get('user')
            if "run_id" in qkit.cfg:
                self.hf.hf.attrs['_run_id'] = qkit.cfg.get('run_id').upper()
        self._mapH5PathToObject()
        self.hf.flush()
        
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.hf.newfile:
            try:
                if type is not None:
                    s = "During creation of this file, the following error ocurred:\n"
                    s += str(exc_type.__name__) + ": " + str(exc_val)+"\n"
                    s += "".join(traceback.format_tb(exc_tb))
                    self.hf.entry.attrs["error"] = s
            except:
                pass
        self.close()

    def _mapH5PathToObject(self):
        """Function for automated data readout at Data object creation.
        
        This function gets called during the init of a data object. Here we 
        translate the attributes of all datasets into attributs of the 
        dummy-class "group". These objects get added to the self.__dict__ what
        makes them tabbable in a notebook.
        We iterate over all atributes of all entries in analysis0 and data0 as
        well as the 'comment' entry on the respective group attributes.
        """
        class group(object):
            pass
        
        a = group()
        for n, o in self.hf.hf['/entry/analysis0'].items():
            n = n.replace(" ","_")
            a.__dict__[n] = o
            for nn, oo in o.attrs.items():
                o.__dict__[nn] = oo
        a.__dict__['comment'] = self.hf.agrp.attrs.get('comment', '')
        d = group()
        for n, o in self.hf.hf['/entry/data0'].items():
            n = n.replace(" ","_")
            d.__dict__[n] = o
            for nn, oo in o.attrs.items():
                o.__dict__[nn] = oo
        d.__dict__['comment'] = self.hf.dgrp.attrs.get('comment', '')
        self.__dict__.update({'analysis':a})
        self.__dict__.update({'data':d})
        v = {a: self.hf.hf.attrs.get(a) for a in self.hf.hf.attrs.keys()}
        del v['NeXus_version'], v['qkit']
        self.__dict__.update(v)

    def _generate_file_name(self):
        """Generate new file name using hdf_DateTimeGenerator.
        
        This sets:
            _unix_timestamp
            _localtime    
            _timestamp
            _timemark
            _datemark
            _uuid
            _filename
            _folder
            _relpath
            _filepath        
        """
        dtg = DateTimeGenerator()
        self.__dict__.update(dtg.new_filename(self._name))
        if self._folder and not os.path.isdir(self._folder):
            os.makedirs(self._folder)

    def __getitem__(self, name):
        return self.hf[name]

    def __setitem__(self, name, val):
        self.hf[name] = val

    def __repr__(self):
        ret = "HDF5Data '%s', filename '%s'" % (self._name, self._filename)
        return ret

    def get_filepath(self):
        return self._filepath

    def get_folder(self):
        return self._folder

    def add_comment(self,comment, folder = "data" ):
        if folder == "data":
            self.hf.dgrp.attrs.create('comment',comment.encode())
        elif folder == "analysis":
            self.hf.agrp.attrs.create("comment",comment.encode())
        else: 
            logging.warning("Foler muset be either 'data' (default) or 'analysis': '%s' provided" % (folder))
            raise ValueError

    def add_textlist(self,name,comment = "" ,folder="data", **meta):
        """Adds a dataset containing only text to the h5 file.
        
        This function is a wrapper to create a hdf_dataset object with some 
        predefined arguments. name, comment, and folder are parsed to the hdf_dataset
        init and the ds_type is set. The dataset does not have any axes.
        
        Args:
            name: String to name the dataset.
            comment: Optional string to put in any comment.
            folder: Optional string ('data' or 'analysis').
        
        Returns:
            hdf_dataset object.
        """
        ds =  hdf_dataset(self.hf, name, comment = comment, folder=folder, ds_type = ds_types['txt'], dim=1, **meta)
        return ds
        
    def add_coordinate(self,  name, unit = "", comment = "",folder="data",**meta):
        """Adds a coordinate dataset to the h5 file.
        
        This function is a wrapper to create a hdf_dataset object with some 
        predefined arguments. name, unit, comment, and folder are parsed to the hdf_dataset
        init and the ds_type is set. The dataset does not have any axes. The data
        is specifically casted to 64bit float.
        
        Args:
            name: String to name the dataset.
            unit: Optional string.
            comment: Optional string to put in any comment.
            folder: Optional string ('data' or 'analysis').
        
        Returns:
            hdf_dataset object.
        """
        ds =  hdf_dataset(self.hf, name,unit=unit, ds_type = ds_types['coordinate'],
                          comment= comment, folder=folder, dtype='float64', dim = 1, **meta)
        return ds

    def add_value_vector(self, name, x, unit = "", comment = "",folder="data",**meta):
        """Adds a 1dim dataset to the h5 file.
        
        This function is a wrapper to create a hdf_dataset object with some 
        predefined arguments. name, x-axis, unit, comment, and folder are parsed
        to the hdf_dataset init and the ds_type is set.
        
        Args:
            name: String to name the dataset.
            x: Optional hdf_dataset representing the x-coordinate of the vector.
            unit: Optional string.
            comment: Optional string to put in any comment.
            folder: Optional string ('data' or 'analysis').
        
        Returns:
            hdf_dataset object.
        """
        ds =  hdf_dataset(self.hf, name, x=x, unit=unit, ds_type = ds_types['vector'],
                          comment=comment, folder=folder, dim = 1, **meta)
        return ds

    def add_value_matrix(self, name, x , y, unit = "", comment = "",folder="data",**meta):
        """Adds a 2dim dataset to the h5 file.
        
        This function is a wrapper to create a hdf_dataset object with some 
        predefined arguments. name, x-axis, y-axis, unit, comment, and folder 
        are parsed to the hdf_dataset init and the ds_type is set.
        Our convention here is: y-axis changes "faster" than the x-axis.
        
        Args:
            name: String to name the dataset.
            x: Optional hdf_dataset representing the x-coordinate of the matrix.
            y: Optional hdf_dataset representing the y-coordinato of the matrix.
            unit: Optional string.
            comment: Optional string to put in any comment.
            folder: Optional string ('data' or 'analysis').
        
        Returns:
            hdf_dataset object.
        """
        ds =  hdf_dataset(self.hf, name, x=x, y=y, unit=unit, ds_type = ds_types['matrix'],
                          comment=comment, folder=folder, dim = 2, **meta)
        return ds

    def add_value_box(self, name, x , y, z, unit = "", comment = "",folder="data",**meta):
        """Adds a 3dim dataset to the h5 file.
        
        This function is a wrapper to create a hdf_dataset object with some 
        predefined arguments. name, x-axis, y-axis, z-axis, unit, comment, and folder 
        are parsed to the hdf_dataset init and the ds_type is set.
        Our convention here is: y-axis changes "faster" than the x-axis.
        
        Args:
            name: String to name the dataset.
            x: Optional hdf_dataset representing the x-coordinate of the box.
            y: Optional hdf_dataset representing the y-coordinato of the box.
            z: Optional hdf_dataset representing the y-coordinato of the box.
            unit: Optional string.
            comment: Optional string to put in any comment.
            folder: Optional string ('data' or 'analysis').
        
        Returns:
            hdf_dataset object.
        """        
        ds =  hdf_dataset(self.hf,name, x=x, y=y, z=z, unit=unit, ds_type = ds_types['box'],
                          comment=comment, folder=folder, dim = 3, **meta)
        return ds

    def add_view(self,name,x = None, y = None, error = None, filter  = None, view_params = {}):
        """Adds a view to plot x-y data.
        
        This function is a wrapper to create a dataset_view.
        x, y with the corresponding error are the datasets to display, e.g.
        x = h5["/entry/data0/temperature"]
        y = h5["/entry/analysis0/frequency_fit"]
        error = h5["/entry/analysis0/frequency_fit_error"]
        filter is a string of reguar python code, which
        accesses the x,y dataset returns arrays of (x,y)
        (Fixme: not jet implemented)
        """
        ds =  dataset_view(self.hf,name, x=x, y=y, error=error, filter = filter, 
                           ds_type = ds_types['view'],view_params = view_params)
        return ds

    def add_fid_param(self, param, value):
        """Adds a parameter (and value) to be read out with the file info database class.
        
        This function adds an entry to the atrribute dict of the self.hf agrp-entry. By default
        this is set to the attribute0 folder. The dict may be read out by the file info database
        class and in the end the entries can be sorted by all params.

        Args:
            param: Parameter name
            value: Parameter value
        """
    
        self.hf.agrp.attrs[param] = value

    def get_dataset(self,ds_url):
        return hdf_dataset(self.hf,ds_url = ds_url)

    def save_finished(self):
        pass

    def flush(self):
        self.hf.flush()

    def close_file(self):
        self.hf.close_file()
    def close(self):
        self.hf.close_file()
